import numpy as np
import ipywidgets as widgets
from IPython.display import display
from ..LinearSystems import interactive_lu


class LUVisualizer:
    def __init__(self, matrix=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float)):
        """
        Parameters
        ----------
        matrix : numpy.ndarray
            Matrix to be decomposed (must be square)

        Returns
        -------
        None

        Exceptions
        ----------
        ValueError
            If the matrix is not square
        """
        self.A = np.array(matrix, dtype=float)
        if self.A.shape[0] != self.A.shape[1]:
            raise ValueError("Matrix must be square")
        self.step = 0
        self.L = np.eye(self.A.shape[0])
        self.U = self.A.copy()
        self.P = np.eye(self.A.shape[0])
        self.rank = 0

        # Stack for previous steps
        self.previous_steps = []

    def initialize_components(self):
        """
        Initialize the components of the visualizer (Output, Buttons, Grid)

        Returns
        -------
        None
        """
        # OUTPUTS
        # ==========================================================================
        ## Output for the matrix P
        self.out_p = widgets.HTMLMath(
            value=pretty_print_matrix(
                self.P, simple=True, type="pMatrix", step=self.step
            ),
            placeholder="$P$",
            description="$P:$",
        )

        ## Output for the matrix L
        self.out_l = widgets.HTMLMath(
            value=pretty_print_matrix(
                self.L, simple=True, type="lMatrix", step=self.step
            ),
            placeholder="$L$",
            description="$\\tilde{L}:$",
        )

        ## Output for the matrix U
        self.out_u = widgets.HTMLMath(
            value=pretty_print_matrix(
                self.U, simple=True, type="uMatrix", step=self.step
            ),
            placeholder="$U$",
            description="$\\tilde{U}:$",
        )

        ## Output for the checker(PA=LU)
        self.out_checker = widgets.HTMLMath(
            value=str(np.allclose(self.L @ self.U, self.P @ self.A)),
            placeholder="$PA=?LU$",
            description="$PA=?LU$",
        )
        ## Output for the message of step skipped normal text
        self.out_message = widgets.Output(
            value="",
            placeholder="Message:",
            description="Message:",
        )
        self.rank_result = widgets.HTMLMath(
            value=f"{self.rank}",
            placeholder="Rank",
            description="Rank:",
        )

        # BUTTONS
        # ==========================================================================
        ## Previous step button
        self.previous_button = widgets.Button(
            description="Previous Step",
            disabled=False,
            button_style="info",
            tooltip="Previous Step",
            icon="arrow-left",
        )
        self.previous_button.on_click(self.previous_step)

        ## Reset button
        self.reset_button = widgets.Button(
            description="Reset",
            disabled=False,
            button_style="danger",
            tooltip="Reset",
            icon="undo",
        )
        self.reset_button.on_click(self.reset)  # Observer for the button

        # MATRIX
        # ==========================================================================
        self.buttons_matrix = []  # List of lists of buttons
        for i in range(self.A.shape[0]):  # For each row
            row = []  # List of buttons
            for j in range(self.A.shape[1]):  # For each column
                row.append(
                    widgets.Button(description=f"{self.A[i,j]:.3f}", disabled=True)
                )  # Create a button with the value of the matrix
                row[-1].index = (i, j)
                # Observer for the button
                row[-1].on_click(self.matrix_pivot_button)

            self.buttons_matrix.append(row)  # Add the row to the list of rows

        self.update_buttons()  # Update the buttons

        # GRID
        # ==========================================================================
        self.grid = widgets.GridspecLayout(6, 5)
        # Add components to grid
        self.grid[0:3, 0:3] = widgets.VBox(  # Add the matrix to the grid
            list(map(lambda x: widgets.HBox(x), self.buttons_matrix))
        )
        self.grid[0, 3] = self.previous_button
        self.grid[1, 3] = self.reset_button
        self.grid[3, 3] = self.rank_result

        self.grid[3, 0] = self.out_p
        self.grid[3, 1] = self.out_l
        self.grid[3, 2] = self.out_u
        self.grid[4, :] = self.out_message

    def matrix_pivot_button(self, b):
        """
        Observer for the buttons in the matrix, when a button is clicked, the pivot is performed and the step is updated

        b.index contains the index of the pivot row
        Call luInteractive with PLU, step and index of pivot row
        Update the output

        Parameters
        ----------
        b : Button
            Button that was clicked

        Returns
        -------
        None
        """
        if b.disabled:
            return  # do nothing if the button is disabled
        self.previous_steps.append(
            (self.P.copy(), self.L.copy(), self.U.copy(), self.step, self.rank)
        )
        with self.grid.hold_sync():  # Hold the sync of the grid
            self.one_step(b.index[0])
            # Update the buttons
            self.update_buttons()

    def one_step(self, pivot):
        # Apply the LU decomposition to the matrix
        self.P, self.L, self.U, self.step, self.rank, msg = interactive_lu(
            self.P, self.L, self.U, self.step, self.rank, pivot
        )
        # Update the output
        self.update_output(msg)
        self.update_buttons()

    def update_buttons(self):
        """
        Update the buttons in the matrix, when a step is performed, blocks the buttons that are not available anymore

        Returns
        -------
        None
        """

        # Update the buttons
        for i in range(len(self.buttons_matrix)):  # For each row
            for j in range(  # For each column
                len(self.buttons_matrix[i])
            ):  # (i,j) is the index of the button
                # Disable the button
                self.buttons_matrix[i][j].disabled = True
                if self.step == -1:  # The end
                    self.buttons_matrix[i][j].style.button_color = "white"
                # Color Green and enable if the button is on the col self.step, the row is greater or eq than the rank and it is not 0
                elif (
                    j == self.step
                    and i >= self.rank
                    and not np.isclose(self.U[i, j], 0)
                ):
                    self.buttons_matrix[i][j].disabled = False
                    self.buttons_matrix[i][j].style.button_color = "LightGreen"
                # Color Red and disable if the button is on the col self.step, the row is greater or eq than the rank and it is 0
                elif j == self.step and i >= self.rank and np.isclose(self.U[i, j], 0):
                    self.buttons_matrix[i][j].style.button_color = "LightCoral"
                # Color white and disable if the button is on the col smaller than the step and the row is smaller than the rank
                elif j < self.step or i < self.rank:
                    self.buttons_matrix[i][j].style.button_color = "LightGray"
                else:  # Color white and disable if the button is on the col greater than the step and the row is greater than the rank
                    self.buttons_matrix[i][j].style.button_color = "white"

                self.buttons_matrix[i][
                    j
                ].style.font_weight = "normal"  # Remove bold from the buttons

                self.buttons_matrix[i][
                    j
                ].description = (
                    f"{self.U[i,j]:.3f}"  # Update the description of the button
                )

    def update_output(self, msg):
        """
        Update the output widgets

        Returns
        -------
        None
        """
        # Update the outputs
        self.out_p.value = pretty_print_matrix(  # Update the output of P
            self.P, simple=True, type="pMatrix", step=self.step
        )
        self.out_l.value = pretty_print_matrix(  # Update the output of L
            self.L, simple=True, type="lMatrix", step=self.step
        )
        self.out_u.value = pretty_print_matrix(  # Update the output of U
            self.U, simple=True, type="uMatrix", step=self.step
        )
        self.rank_result.value = (  # Update the output of the rank
            f"${self.rank}$ "
            if self.step != -1
            else f"$\\underline{{\\textbf{{{self.rank}}}}}$"
        )

        self.out_message.clear_output()  # Clear the output of the messages
        with self.out_message:  # Print the messages
            if msg != "":
                print(
                    "Messages from system: " + msg
                )  # Print the message if it is not empty
            else:
                print("")  # Print an empty line if the message is empty

    def previous_step(self, b):
        """
        Observer for the previous step button, when clicked, it returns the state to the previous step

        Parameters
        ----------
        b : Button
            Button that was clicked (Not used)

        Returns
        -------
        None
        """
        # If there are previous steps, then go back to the previous step
        if len(self.previous_steps) > 0:  # If there are previous steps in the stack
            (
                self.P,
                self.L,
                self.U,
                self.step,
                self.rank,
            ) = self.previous_steps.pop()  # Pop the previous step from the stack
            with self.grid.hold_sync():  # Hold the sync of the grid
                self.update_output("")  # Update the output
                # Update the buttons
                self.update_buttons()  # Update the buttons

    def reset(self, b):
        """
        Observer for the reset button, when clicked, it returns the state to the initial state

        Parameters
        ----------
        b : Button
            Button that was clicked (Not used)

        Returns
        -------
        None
        """

        # Reset the LU decomposition Visualizer to the initial state
        self.step = 0  # Set the step to 0
        self.rank = 0  # Set the rank to 0
        self.L = np.eye(self.A.shape[0])  # Set L to the identity matrix
        self.U = self.A.copy()  # Set U to the matrix A
        self.P = np.eye(self.A.shape[0])  # Set P to the identity matrix

        # Stack for previous steps
        self.previous_steps = []  # Clear the stack of previous steps
        with self.grid.hold_sync():  # Hold the sync of the grid
            # Update the output
            self.update_output("")  # Update the output
            # Update the buttons
            self.update_buttons()  # Update the buttons

    def run(self):
        """
        Run the LU decomposition Visualizer

        Returns
        -------
        None
        """
        # Run the visualizer
        self.initialize_components()
        # if the first column is all 0, then automatically perform the first step
        if np.all(self.U[:, 0] == 0):
            # print("First column is all 0, automatically performing the first step")
            with self.grid.hold_sync():
                self.one_step(0)

        return self.grid


def pretty_print_matrix(matrix, simple=False, type="normal", step=0):
    res = "  \\begin{pmatrix} \n"

    if type in ["normal", "pMatrix"]:
        for (
            row
        ) in matrix:  # Corresponds to any matrix or the P matrix (Permutation matrix)
            res += " & ".join([str(round(x, 3)) for x in row]) + "\\\\ \n"
    elif type == "lMatrix":  # Corresponds to the L matrix (Lower triangular matrix)
        # Write * everywhere except for the diagonal and the colums before the step
        for i in range(len(matrix)):
            for j in range(len(matrix[i])):
                if (i <= j or j < step) or step == -1:
                    res += str(round(matrix[i][j], 3)) + " & "
                else:
                    res += "* & "
            res = res[:-2] + "\\\\ \n"
    elif type == "uMatrix":  # Corresponds to the U matrix (Upper triangular matrix)
        # Write * on the submatrix [row:, step:]
        for i in range(len(matrix)):
            for j in range(len(matrix[i])):
                if i < step or j < step or step == -1:
                    res += str(round(matrix[i][j], 3)) + " & "
                else:
                    res += "* & "
            res = res[:-2] + "\\\\ \n"

    res += "\\end{pmatrix} "

    return display.Math(res) if not simple else res



from unittest import TestCase
from BNumMet.module import pretty_print_matrix, sort_interpolation_values, pretty_plua
from BNumMet.LinearSystems import lu
import numpy as np


class test_module(TestCase):
    def test_PrettyPrint(self):
        matrix = [[1, 2], [3, 4]]
        stringRes = """ \\begin{pmatrix}1 & 2\\\\3 & 4\\\\\\end{pmatrix}"""

        self.assertTrue(pretty_print_matrix(matrix).replace("\n", "") == stringRes)

    def test_SortInterpolVals(self):
        x = [1, 3, 2]
        y = [4, 3, 2]
        xRes = [1, 2, 3]
        yRes = [4, 2, 3]
        x, y = sort_interpolation_values(x, y)

        self.assertTrue((x == xRes).all() and (y == yRes).all())

    def test_pretty_plua(self):
        matrix = [[1, 2], [3, 4]]

        p, l, u = lu(np.array(matrix))
        res = pretty_plua(p, l, u, matrix)
        # assert type of p, l, u, a is str
        self.assertTrue(type(res) == str)



import numpy as np


def pretty_print_matrix(matrix):
    """
    Prints a matrix in a pretty format

    Parameters
    ----------
    matrix : list
        The matrix to print

    Returns
    -------
    res : str
        The string representation of the matrix in a pretty format for LaTeX
    """
    # Initialize the string to represent the matrix
    res = " \\begin{pmatrix}\n"

    # Loop through each row in the matrix
    for row in matrix:
        # Join the elements in each row with " & " and add a line break at the end
        res += " & ".join([str(x) for x in row]) + "\\\\\n"

    # Close the matrix representation
    res += "\end{pmatrix}"

    # Return the final string
    return res


def sort_interpolation_values(x, y):
    """
    Sorts the interpolation values by the x coordinates

    Parameters
    ----------
    x : list
        The x coordinates of the interpolation values
    y : list
        The y coordinates of the interpolation values

    Returns
    -------
    x : list
        The sorted x coordinates of the interpolation values
    y : list
        The sorted y coordinates of the interpolation values
    """
    x = np.array(x)  # Convert the x coordinates to a numpy array
    y = np.array(y)  # Convert the y coordinates to a numpy array
    ind = np.argsort(x)  # Get the indices of the sorted array
    x = x[ind]  # Sort the x coordinates
    y = y[ind]  # Sort the y coordinates

    return x, y


def pretty_plua(P, L, U, A):
    res = f""" 
    \\begin{{array}}{{lll}}
        P = {pretty_print_matrix(P)} & L = {pretty_print_matrix(L)} & U = {pretty_print_matrix(U)} 
        \\\\
        &&
        \\\\
        P\cdot A = {pretty_print_matrix(P@A)} &  L\cdot U = {pretty_print_matrix(L@U)}
    \\end{{array}}
    """
    return res



import os
import sys
import unittest
from random import randint
from unittest import TestCase

import numpy as np
import pytest

from BNumMet.LinearSystems import (
    backward_substitution,
    forward_substitution,
    interactive_lu,
    lu,
    lu_solve,
    permute,
    qr_factorization,
    qr_solve,
)
from BNumMet.Visualizers.LUVisualizer import LUVisualizer


class test_LU(TestCase):
    def test_lu_simple(self):
        """
        Test the LU decomposition by running it on a fixed matrix and checking that the result is correct
        """
        A = np.array([[10, -7, 0], [-3, 2, 6], [5, -1, 5]])
        P, L, U = lu(A)
        # print(P@A)
        print(P)
        print(L)
        print(U)

        self.assertTrue(np.allclose(P @ A, L @ U))

    def test_lu_random(self):
        """
        Test the LU decomposition by running it on a random matrix and checking that the result is correct
        """
        A = np.random.rand(10, 10)
        P, L, U = lu(A)

        self.assertTrue(np.allclose(P @ A, L @ U), msg=f"P@A != L@U\n{P@A} != {L@U}")

    def test_lu_random_ints(self):
        """
        Test the LU decomposition by running it on a random INTEGER matrix and checking that the result is correct
        This test is useful to check that the algorithm works with integer matrices as well because the python maintains the type of the elements of the matrix when performing operations
        """
        A = np.random.randint(0, 10, (10, 10))
        P, L, U = lu(A)

        self.assertTrue(np.allclose(P @ A, L @ U), msg=f"P@A != L@U\n{P@A} != {L@U}")

    def test_lu_notSquare(self):
        """
        Test that the LU decomposition raises a ValueError when the matrix is not square
        """
        A = np.array([[10, -7, 0], [-3, 2, 6], [5, -1, 5], [1, 2, 3]])
        with self.assertRaises(ValueError):
            P, L, U = lu(A)

    def test_permute(self):
        """
        Test the permute function by running it on a fixed matrix and checking that the result is correct
        It also checks that the function does not modify the original matrix
        """
        A = np.array([[10, -7, 0], [-3, 2, 6], [5, -1, 5]])

        for i in range(A.shape[0]):
            self.assertTrue(
                np.allclose(permute(A, i, i), A)
            )  # Check trivial permutation
        self.assertTrue(
            np.allclose(
                permute(A, 0, 1), np.array([[-3, 2, 6], [10, -7, 0], [5, -1, 5]])
            )
        )  # swap 1st and 2nd rows
        self.assertFalse(
            np.allclose(A, np.array([[10, -7, 0], [-3, 2, 6], [5, -1, 5]]))
        )  # It is not the same matrix as A - initially

    def test_interactive_lu(self):
        """
        Test the interactive LU decomposition by running it on a fixed matrix and checking that the result is correct
        we do not test the visualizer here, just the algorithm, with automated pivoting (i.e. iMax = -1)
        """
        A = np.array([[10, -7, 0], [-3, 2, 6], [5, -1, 5]])
        L = np.eye(A.shape[0])
        U = A.copy()
        P = np.eye(A.shape[0])

        lastColumn = 0
        rank = 0
        while lastColumn != -1:
            P, L, U, lastColumn, rank, msg = interactive_lu(
                P, L, U, lastColumn, rank, -1
            )
            self.assertTrue(np.allclose(P @ A, L @ U))
        print(rank)
        assert rank == A.shape[0]

    def test_interactive_lu_ranks(self):
        """
        Test the interactive LU decomposition by running it on a fixed matrix and checking that the result is correct
        we do not test the visualizer here, just the algorithm, with automated pivoting (i.e. iMax = -1)
        """
        A = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        L = np.eye(A.shape[0])
        U = A.copy()
        P = np.eye(A.shape[0])

        _, _, _, _, rank, _ = interactive_lu(P, L, U, 0, 0, -1)
        self.assertTrue(rank == 0)

        A = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        L = np.eye(A.shape[0])
        U = A.copy()
        P = np.eye(A.shape[0])

        _, _, _, _, rank, _ = interactive_lu(P, L, U, 0, 0, -1)
        col = 0
        rank = 0
        while col != -1:
            P, L, U, col, rank, msg = interactive_lu(P, L, U, col, rank, -1)
            self.assertTrue(np.allclose(P @ A, L @ U))
        self.assertTrue(rank == A.shape[0])

        A = np.array([[1, 4, 7], [1, 5, 7], [1, 4, 7]])  # rank 2 on the second column
        L = np.eye(A.shape[0])
        U = A.copy()
        P = np.eye(A.shape[0])
        col = 0
        rank = 0
        while col != -1:
            P, L, U, col, rank, msg = interactive_lu(P, L, U, col, rank, -1)
            self.assertTrue(np.allclose(P @ A, L @ U))

        self.assertTrue(rank == 2)

        A = np.array([[1, 4, 7], [1, 4, 7], [1, 4, 9]])  # rank 2 on the third column
        L = np.eye(A.shape[0])
        U = A.copy()
        P = np.eye(A.shape[0])
        col = 0
        rank = 0
        while col != -1:
            P, L, U, col, rank, msg = interactive_lu(P, L, U, col, rank, -1)
            self.assertTrue(np.allclose(P @ A, L @ U))

        self.assertTrue(rank == 2)  #

    def test_forwardSubstitution(self):
        """
        Test the forward substitution algorithm by running it on a fixed matrix and checking that the result is correct, testing BNumMet.forwardSubstitution
        """
        L = np.array([[1, 0, 0], [0.5, 1, 0], [0.5, 0.5, 1]])
        b = np.array([1, 2, 3])
        x = np.array([1, 1.5, 1.75])

        self.assertTrue(np.allclose(forward_substitution(L, b), x))

    def test_backwardSubstitution(self):
        """
        Test the backward substitution algorithm by running it on a fixed matrix and checking that the result is correct, testing BNumMet.backwardSubstitution
        """
        U = np.array([[1, 2, 3], [0, 1, 2], [0, 0, 1]])
        b = np.array([1, 2, 3])
        x = np.array([0, -4, 3])

        self.assertTrue(np.allclose(backward_substitution(U, b), x))

    def test_forwardSubstitution_expceptions(self):
        """
        Test the forward substitution algorithm by running it on a fixed matrix and checking that the result is correct, testing BNumMet.forwardSubstitution
        """
        L = np.array([[1, 0, 0], [0.5, 1, 0], [0.5, 0.5, 1]])
        b = np.array([1, 2, 3, 4])
        x = np.array([1, 1.5, 2])
        # 1. The left hand side has more rows than the right hand side
        with self.assertRaises(ValueError, msg="b has more elements than L"):
            forward_substitution(L, b)  # b has more elements than L

        # 2. The left hand side has less rows than the right hand side
        L = np.array([[1, 0, 0], [0.5, 1, 0], [0.5, 0.5, 1]])
        b = np.array([1, 2])
        x = np.array([1, 1.5, 2])

        with self.assertRaises(ValueError, msg="L has more rows than b"):
            forward_substitution(L, b)

        # 3. L is not lower triangular
        L = np.array([[1, 1, 0], [0.5, 1, 0], [0.5, 0.5, 1]])
        b = np.array([1, 2, 3])
        x = np.array([1, 1.5, 2])

        with self.assertRaises(ValueError, msg="L is not lower triangular"):
            forward_substitution(L, b)

        # 4. L has a zero diagonal element
        L = np.array([[1, 0, 0], [0.5, 0, 0], [0.5, 0.5, 1]])
        b = np.array([1, 2, 3])
        x = np.array([1, 1.5, 2])
        with self.assertRaises(ValueError, msg="L has a zero diagonal element"):
            forward_substitution(L, b)

        # 5. L is not a square matrix
        L = np.array([[1, 0, 0], [0.5, 1, 0], [0.5, 0.5, 1], [0.5, 0.5, 1]])
        b = np.array([1, 2, 3, 4])
        x = np.array([1, 1.5, 2])
        with self.assertRaises(ValueError, msg="U is not a square matrix") as e:
            forward_substitution(L, b)
        # Check if e message is correct
        self.assertEqual(
            str(e.exception),
            "A is not a square matrix",
            msg="The error message is not correct",
        )

    def test_backwardSubstitution_expceptions(self):
        """
        Test the backward substitution algorithm by running it on a fixed matrix and checking that the result is correct, testing BNumMet.backwardSubstitution
        """
        U = np.array([[1, 2, 3], [0, 1, 2], [0, 0, 1]])
        b = np.array([1, 2, 3, 4])
        x = np.array([1, 1.5, 2])
        # 1. The left hand side has more rows than the right hand side
        with self.assertRaises(ValueError, msg="b has more elements than U"):
            backward_substitution(U, b)  # b has more elements than L

        # 2. The left hand side has less rows than the right hand side
        U = np.array([[1, 2, 3], [0, 1, 2], [0, 0, 1]])
        b = np.array([1, 2])
        x = np.array([1, 1.5, 2])

        with self.assertRaises(ValueError, msg="U has more rows than b"):
            backward_substitution(U, b)

        # 3. U is not upper triangular
        U = np.array([[1, 2, 3], [0, 1, 2], [0, 1, 1]])
        b = np.array([1, 2, 3])
        x = np.array([1, 1.5, 2])

        with self.assertRaises(ValueError, msg="U is not upper triangular"):
            backward_substitution(U, b)

        # 4. U has a zero diagonal element
        U = np.array([[1, 2, 3], [0, 0, 2], [0, 0, 1]])
        b = np.array([1, 2, 3])
        x = np.array([1, 1.5, 2])

        with self.assertRaises(ValueError, msg="U has a zero diagonal element"):
            backward_substitution(U, b)

        # 5. U is not a square matrix
        U = np.array([[1, 2, 3], [0, 1, 2], [0, 0, 1], [0, 0, 0]])
        b = np.array([1, 2, 3, 4])
        x = np.array([1, 1.5, 2])

        with self.assertRaises(ValueError, msg="U is not a square matrix") as e:
            backward_substitution(U, b)
        # Check if e message is correct
        self.assertEqual(
            str(e.exception),
            "A is not a square matrix",
            msg="The error message is not correct",
        )

    def test_luSolve(self):
        """
        Test the LU decomposition algorithm by running it on a fixed matrix and checking that the result is correct, testing BNumMet.luSolve
        """
        A = np.array([[1, 1], [1, -1]])
        b = np.array([10, 4])
        x = np.array([7, 3])

        solution = lu_solve(A, b)

        self.assertTrue(
            np.allclose(solution, x),
            msg=f"The solution is not correct {solution} != {x}",
        )

        A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        b = np.array([1, 2, 3])

        with self.assertRaises(
            ValueError, msg="A is Singular but it does not detect it!"
        ):
            lu_solve(A, b)


class Test_LinearSystems(TestCase):
    def test_qrFactorization(self):
        """
        Test the QR decomposition algorithm by running it on a fixed matrix and checking that the result is correct, testing BNumMet.qr_factorization
        """
        # Test 1: Simple 2x2 matrix
        A = np.array([[1, 2], [3, 4]])
        Q, R = qr_factorization(A)
        self.assertTrue(
            np.all(np.tril(R, -1) < 1e-15), f"R is not upper triangular {np.tril(R,-1)}"
        )
        self.assertTrue(np.allclose(A, Q @ R))

        # Test 2: Simple 3x3 matrix
        A = np.array([[1, 2, 3], [4, 8, 6], [7, 8, 9]])
        Q, R = qr_factorization(A)
        self.assertTrue(np.allclose(A, Q @ R), f"{A} != {Q @ R}")

        # Test 3: Simple 4x3 matrix
        A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        Q, R = qr_factorization(A)
        self.assertTrue(np.allclose(A, Q @ R))

    def test_qrSolve(self):
        # Test case 1: Solve a 2x2 system
        A = np.array([[1, 1], [1, -1]])
        b = np.array([10, 4])
        x = np.array([7, 3])

        self.assertTrue(np.allclose(qr_solve(A, b.T), x), f"{x} != {qr_solve(A, b)}")

        # Test case 2: Solve a 3x3 system
        A = np.array([[2, -1, 0], [1, 2, -1], [0, 1, 2]])
        b = np.array([0, 2, 8])
        x = np.array([1, 2, 3])

        self.assertTrue(np.allclose(qr_solve(A, b), x), f"{x} != {qr_solve(A, b)}")

        # Test case 3: Solve a 4x4 system
        A = np.array([[2, 3, -1, 1], [1, -2, 0, -1], [2, -1, 3, -2], [1, 1, 1, 2]])
        x = np.array([2, -1, 3, 1])
        b = A @ x

        self.assertTrue(np.allclose(qr_solve(A, b), x), f"{x} != {qr_solve(A, b)}")

    def test_qrSolve_random(self):
        """
        Test the QR decomposition algorithm by running it on a random matrix and checking that the result is correct, testing BNumMet.qrSolve
        """
        for i in range(10):
            size = np.random.randint(2, 15)
            A = np.random.rand(size, size)
            x = np.ones(size)
            b = A @ x

            self.assertTrue(np.allclose(qr_solve(A, b), x), f"{x} != {qr_solve(A, b)}")

    def test_qr_lsp(self):
        """
        Test the QR solver for an mxn matrix where m > n
        """
        A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        b = np.array([1, 2, 3, 4])
        x = np.array([1, 1, 1])

        # no exception should be raised
        qr_solve(A, b)


class Test_LUVisualizer(TestCase):
    def runtest_setup(self, A=None):
        """
        Run the setup for the tests of the LUVisualizer

        Parameters
        ----------
        A : np.array
            The matrix to be decomposed. If None, a fixed matrix is put
        """
        self.A = (
            np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float) if A is None else A
        )
        self.luVisualizer = LUVisualizer(self.A)
        self.luVisualizer.run()

    def test_no_param_init(self):
        """
        Test if the LUVisualizer is initialized correctly when no matrix is passed as a parameter to the constructor (i.e. A default matrix is used)
        """
        A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float)
        self.luVisualizer = LUVisualizer()
        self.luVisualizer.run()

        self.assertTrue(np.allclose(self.luVisualizer.A, A))

    def test_initialazation(self):
        """
        Test if the LUVisualizer is initialized correctly
        """
        self.runtest_setup()
        self.assertEqual(self.luVisualizer.step, 0)
        self.assertTrue(np.allclose(self.A, self.luVisualizer.A))

        self.assertTrue(self.A.shape == self.luVisualizer.P.shape)
        self.assertTrue(self.A.shape == self.luVisualizer.L.shape)

    def test_ValueError(self):
        """
        Test if the LUVisualizer raises a ValueError when the matrix is not square
        """
        with self.assertRaises(ValueError):
            self.luVisualizer = LUVisualizer([[1, 2, 3, 4], [4, 5, 6, 4], [7, 8, 9, 4]])

    def test_step(self):
        """
        Test if step works correctly by clicking one button and then checking if the step has changed, and if the buttons have changed too
        """
        self.runtest_setup()
        # Get buttons which are not disabled
        buttons = [
            button
            for row in self.luVisualizer.buttons_matrix
            for button in row
            if not button.disabled
        ]

        # Simulate clicking one of the buttons
        buttons[0].click()

        # This should hgave changed the step and the matrix buttons
        self.assertEqual(self.luVisualizer.step, 1)
        buttons2 = [
            button
            for row in self.luVisualizer.buttons_matrix
            for button in row
            if not button.disabled
        ]
        self.assertNotEqual(buttons, buttons2)

        # The button should be disabled
        self.assertTrue(buttons[0].disabled)
        buttons[0].click()
        self.assertEqual(self.luVisualizer.step, 1)

        buttons2[0].click()
        self.assertEqual(
            self.luVisualizer.step, -1
        )  # The last is a Linear Combination --> Step +1 but it is the last step --> -1

    def test_previousStep(self):
        """
        Test if previousStep works correctly by clicking one button and then going back
        """
        self.runtest_setup()

        oldStep = (
            self.luVisualizer.step,
            self.luVisualizer.L,
            self.luVisualizer.U,
            self.luVisualizer.P,
        )

        self.assertTrue(len(self.luVisualizer.previous_steps) == 0)
        # Get buttons which are not disabled
        buttons = [
            button
            for row in self.luVisualizer.buttons_matrix
            for button in row
            if not button.disabled
        ]
        # Simulate clicking one of the buttons
        buttons[0].click()
        self.assertTrue(len(self.luVisualizer.previous_steps) == 1)

        self.luVisualizer.previous_step(None)
        self.assertEqual(self.luVisualizer.step, oldStep[0])
        self.assertTrue(np.allclose(self.luVisualizer.L, oldStep[1]))
        self.assertTrue(np.allclose(self.luVisualizer.U, oldStep[2]))
        self.assertTrue(np.allclose(self.luVisualizer.P, oldStep[3]))

    def test_reset(self):
        """
        Test reset button by clicking all buttons and then reset
        It should reset the step and the matrices to the initial values
        """
        self.runtest_setup()

        buttons = [
            button
            for row in self.luVisualizer.buttons_matrix
            for button in row
            if not button.disabled
        ]
        while len(buttons) > 0:
            buttons[0].click()
            buttons = [
                button
                for row in self.luVisualizer.buttons_matrix
                for button in row
                if not button.disabled
            ]

        # All buttons should be disabled
        self.assertTrue(
            all(
                [
                    button.disabled
                    for row in self.luVisualizer.buttons_matrix
                    for button in row
                ]
            )
        )

        self.luVisualizer.reset(None)
        self.assertEqual(self.luVisualizer.step, 0)
        self.assertTrue(np.allclose(self.luVisualizer.L, np.eye(self.A.shape[0])))
        self.assertTrue(np.allclose(self.luVisualizer.U, self.A))
        self.assertTrue(np.allclose(self.luVisualizer.P, np.eye(self.A.shape[0])))
        self.assertTrue(
            any(
                [
                    not button.disabled
                    for row in self.luVisualizer.buttons_matrix
                    for button in row
                ]
            )
        )

    def test_interactive_lu_proccess(self):
        """
        Test the interactive LU proccess:
        1. Generate a random matrix A (10x10)
        2. Run the interactive LU proccess

        for each step:
            3. Click a random button which is not disabled
            4. Check that P@A = L@U
        """
        self.runtest_setup(np.random.rand(10, 10))
        P = self.luVisualizer.P
        L = self.luVisualizer.L
        U = self.luVisualizer.U
        A = self.luVisualizer.A
        for i in range(self.A.shape[0]):
            buttons = [
                button
                for row in self.luVisualizer.buttons_matrix
                for button in row
                if not button.disabled
            ]
            if len(buttons) == 0:
                break
            buttons[randint(0, len(buttons) - 1)].click()

            P = self.luVisualizer.P
            L = self.luVisualizer.L
            U = self.luVisualizer.U
            self.assertTrue(
                np.allclose(
                    P @ A,
                    L @ U,
                ),
                msg=f"P@A != L@U\n{P@A} != {L@U}",
            )

        endButtons = self.luVisualizer.buttons_matrix
        for row in endButtons:
            for button in row:
                self.assertTrue(button.disabled)



## Lehmers Random Number Generator
lehmers_vars = {  # Xn+1 = a*Xn + c mod m
    "a": None,  # Multiplier
    "c": None,  # Increment
    "m": None,  # Modulus
    "x": None,  # Seed/Current Value
}


def clear_lehmers_vars():
    """
    This function clears the global dictionary lehmersVars
    """
    global lehmers_vars
    lehmers_vars = {  # Xn+1 = a*Xn + c mod m
        "a": None,  # Multiplier
        "c": None,  # Increment
        "m": None,  # Modulus
        "x": None,  # Seed/Current Value
    }


def lehmers_init(a, c, m, x):
    """
    This function initializes the parameters for the Lehmer's Random Number Generator

    Parameters
    ----------
    a : int
        The multiplier
    c : int
        The increment
    m : int
        The modulus
    x : int
        The seed

    Returns
    -------
    None
    """
    global lehmers_vars
    # Set the values for the global dictionary lehmersVars
    lehmers_vars["a"] = a
    lehmers_vars["c"] = c
    lehmers_vars["m"] = m
    lehmers_vars["x"] = x


def lehmers_rand(a=None, c=None, m=None, x=None):
    """
    This function generates a random number using Lehmer's Random Number Generator. It takes in 4 optional parameters a, c, m, x which can be used to initialize the generator

    Parameters
    ----------
    a : int, optional
        The multiplier
    c : int, optional
        The increment
    m : int, optional
        The modulus
    x : int, optional
        The seed

    Returns
    -------
    float
        A random number between 0 and 1
    """
    global lehmers_vars
    # Check if the generator has been initialized, if not, use default values to initialize it
    if (
        lehmers_vars["a"] == None
        or lehmers_vars["c"] == None
        or lehmers_vars["m"] == None
        or lehmers_vars["x"] == None
    ):
        if a == None or c == None or m == None or x == None:
            # Initialize the generator with default values
            lehmers_init(7**5, 0, 2**31 - 1, 1.0)
            # Print out the initialized values
            print(
                f'Lehmers Random Number Generator Initialized with default values\n\ta={lehmers_vars["a"]}\n\tc={lehmers_vars["c"]}\n\tm={lehmers_vars["m"]}\n\tx={lehmers_vars["x"]}'
            )
        else:
            # Initialize the generator with the given values
            lehmers_init(a, c, m, x)
    # Generate a random number using Lehmer's formula
    lehmers_vars["x"] = (
        lehmers_vars["a"] * lehmers_vars["x"] + lehmers_vars["c"]
    ) % lehmers_vars["m"]
    # Return the generated random number
    return lehmers_vars["x"] / lehmers_vars["m"]


# Marsaglia's Random Number Generator: Subtract with Borrow
marsaglia_vars = {  # f(x_1...x_r,c) =    | (x_2,...,x_r, x_1-x_{r-s} - c , 0)      if x_1-x_{r-s} - c >= 0
    #                     | (x_2,...,x_r, x_1-x_{r-s} - c + b, 1)   if x_1-x_{r-s} - c < 0
    "base": None,  # Base
    "lag_r": None,  # Lag r
    "lag_s": None,  # Lag s
    "carry": None,  # Carry
    "args": None,  # Values of x_1...x_r
}


def clear_marsaglia_vars():
    """
    This function clears the global dictionary marsagliaVars
    """
    global marsaglia_vars
    marsaglia_vars = {  # f(x_1...x_r,c) =    | (x_2,...,x_r, x_1-x_{r-s} - c , 0)      if x_1-x_{r-s} - c >= 0
        #                     | (x_2,...,x_r, x_1-x_{r-s} - c + b, 1)   if x_1-x_{r-s} - c < 0
        "base": None,  # Base
        "lag_r": None,  # Lag r
        "lag_s": None,  # Lag s
        "carry": None,  # Carry
        "args": None,  # Values of x_1...x_r
    }


def marsaglia_init(base, lag_r, lag_s, carry, seed_tuple):
    """
    This function initializes the parameters for the Marsaglia's Random Number Generator

    Parameters
    ----------
    base : int
        The base
    lag_r : int
        The lag r
    lag_s : int
        The lag s
    carry : int
        The carry
    seed_tuple : tuple
        The seed tuple

    Returns
    -------
    None
    """

    global marsaglia_vars
    if (
        type(seed_tuple) == tuple and len(list(seed_tuple)) != 2
    ):  # Check if the seed_tuple is of length 2
        raise ValueError("seed_tuple must be a tuple of length 2")
    elif type(seed_tuple) != tuple:  # Check if the seed_tuple is a tuple
        raise ValueError("seed_tuple must be a tuple ")
    if lag_r < 1 or lag_s < 1:  # Check if lag_r and lag_s are greater than 0
        raise ValueError("lag_r and lag_s must be greater than 0")
    if lag_r < lag_s:  # Check if lag_r is greater than or equal to lag_s
        raise ValueError("lag_r must be greater than or equal to lag_s")
    if carry < 0 or carry > 1:  # Check if carry is 0 or 1
        raise ValueError("carry must be 0 or 1")
    if base < 1:  # Check if base is greater than 0
        raise ValueError("base must be greater than 0")
    # Set the values for the global dictionary marsagliaVars
    marsaglia_vars["base"] = base
    marsaglia_vars["lag_r"] = lag_r
    marsaglia_vars["lag_s"] = lag_s
    marsaglia_vars["carry"] = carry
    marsaglia_vars["args"] = [0] * lag_r
    marsaglia_vars["args"][0] = seed_tuple[
        0
    ]  # Set the value of x_1 to the first element of the seed_tuple
    marsaglia_vars["args"][lag_r - lag_s] = seed_tuple[
        1
    ]  # Set the value of x_{r-s} to the second element of the seed_tuple


def marsaglia_rand(base=None, lag_r=None, lag_s=None, carry=None, seed_tuple=None):
    """
    This function generates a random number using Marsaglia's Random Number Generator

    Parameters
    ----------
    base : int, optional
        The base
    lag_r : int, optional
        The lag r
    lag_s : int, optional
        The lag s
    carry : int, optional
        The carry
    seed_tuple : tuple, optional
        The seed tuple

    Returns
    -------
    float
        The generated random number
    """
    global marsaglia_vars  # Use the global dictionary marsagliaVars
    if (  # Check if the global dictionary marsagliaVars is initialized
        marsaglia_vars["base"] == None
        or marsaglia_vars["lag_r"] == None
        or marsaglia_vars["lag_s"] == None
        or marsaglia_vars["carry"] == None
        or marsaglia_vars["args"] == None
    ):
        if (  # Check if the parameters are None
            base == None
            or lag_r == None
            or lag_s == None
            or carry == None
            or seed_tuple == None
        ):
            marsaglia_init(
                2**31 - 1, 19, 7, 1, (1, 1)
            )  # Initialize the global dictionary marsagliaVars with default values
            print(
                f'Marsaglia Random Number Generator Initialized with default values\n\tbase={marsaglia_vars["base"]}\n\tlag_r={marsaglia_vars["lag_r"]}\n\tlag_s={marsaglia_vars["lag_s"]}\n\tcarry={marsaglia_vars["carry"]}\n\tseed_tuple={marsaglia_vars["args"]}'
            )
        else:  # Initialize the global dictionary marsagliaVars with the parameters passed
            marsaglia_init(base, lag_r, lag_s, carry, seed_tuple)
    new_random_number = (
        marsaglia_vars["args"][0]
        - marsaglia_vars["args"][marsaglia_vars["lag_r"] - marsaglia_vars["lag_s"]]
        - marsaglia_vars["carry"]
    )  # f(x_1...x_r,c) =    | (x_2,...,x_r, x_1-x_{r-s} - c , 0)      if x_1-x_{r-s} - c >= 0
    #                     | (x_2,...,x_r, x_1-x_{r-s} - c + b, 1)   if x_1-x_{r-s} - c < 0
    if new_random_number < 0:  # Check if x_1-x_{r-s} - c < 0
        new_random_number += marsaglia_vars["base"]  # Add b to x_1-x_{r-s} - c
        marsaglia_vars["carry"] = 1  # Set carry to 1
    else:
        marsaglia_vars["carry"] = 0  # Set carry to 0

    marsaglia_vars["args"] = marsaglia_vars["args"][1:] + [
        new_random_number
    ]  # Set the values of x_1...x_r to x_2...x_r, x_1-x_{r-s} - c
    return new_random_number / marsaglia_vars["base"]  # Return the random number


# Mersenne Twister Random Number Generator
mt_vars = {
    "N": None,  # N
    "M": None,  # M
    "MATRIX_A": None,  # MATRIX_A
    "UPPER_MASK": None,  # UPPER_MASK
    "LOWER_MASK": None,  # LOWER_MASK
    "TEMPERING_MASK_B": None,  # TEMPERING_MASK_B
    "TEMPERING_MASK_C": None,  # TEMPERING_MASK_C
    "mt": None,  # mt
    "mti": None,  # mti
}


def clear_mt_vars():
    """
    This function clears the global dictionary mtVars

    Returns
    -------
    None
    """
    global mt_vars
    mt_vars = {
        "N": None,  # N
        "M": None,  # M
        "MATRIX_A": None,  # MATRIX_A
        "UPPER_MASK": None,  # UPPER_MASK
        "LOWER_MASK": None,  # LOWER_MASK
        "TEMPERING_MASK_B": None,  # TEMPERING_MASK_B
        "TEMPERING_MASK_C": None,  # TEMPERING_MASK_C
        "mt": None,  # mt
        "mti": None,  # mti
    }


def sgenrand(seed):
    """
    This function initializes the global dictionary mtVars

    Parameters
    ----------
    seed : int
        The seed

    Returns
    -------
    None
    """
    global mt_vars
    mt_vars = {
        "N": 624,  # N
        "M": 397,  # M
        "MATRIX_A": 0x9908B0DF,  # MATRIX_A
        "UPPER_MASK": 0x80000000,  # UPPER_MASK
        "LOWER_MASK": 0x7FFFFFFF,  # LOWER_MASK
        "TEMPERING_MASK_B": 0x9D2C5680,  # TEMPERING_MASK_B
        "TEMPERING_MASK_C": 0xEFC60000,  # TEMPERING_MASK_C
        "mt": [0] * 624,  # mt
        "mti": 624 + 1,  # mti
        "TEMPERING_SHIFT_U": lambda y: y >> 11,  # TEMPERING_SHIFT_U(y)
        "TEMPERING_SHIFT_S": lambda y: y << 7,  # TEMPERING_SHIFT_S(y)
        "TEMPERING_SHIFT_T": lambda y: y << 15,  # TEMPERING_SHIFT_T(y)
        "TEMPERING_SHIFT_L": lambda y: y >> 18,  # TEMPERING_SHIFT_L(y)
    }

    mt_vars["mt"][0] = seed & 0xFFFFFFFF
    mt_vars["mti"] = 1
    while mt_vars["mti"] < mt_vars["N"]:
        mt_vars["mt"][mt_vars["mti"]] = (
            69069 * mt_vars["mt"][mt_vars["mti"] - 1]
        ) & 0xFFFFFFFF
        mt_vars["mti"] += 1


def genrand(seed: int = None):
    global mt_vars

    # Check if the global dictionary mtVars is initialized
    if (
        mt_vars["N"] is None
        or mt_vars["M"] is None
        or mt_vars["MATRIX_A"] is None
        or mt_vars["UPPER_MASK"] is None
        or mt_vars["LOWER_MASK"] is None
        or mt_vars["TEMPERING_MASK_B"] is None
        or mt_vars["TEMPERING_MASK_C"] is None
        or mt_vars["mt"] is None
        or mt_vars["mti"] is None
    ):
        if seed is None:
            seed = 4357
        if type(seed) is not int:
            raise ValueError("The seed must be an integer")
        sgenrand(seed)
        print("Initialized the global dictionary mtVars with seed " + str(seed))

    mag01 = [0x0, mt_vars["MATRIX_A"]]

    # Generate N words at one time
    if mt_vars["mti"] >= mt_vars["N"]:
        kk = 0
        y = None

        while kk < mt_vars["N"] - mt_vars["M"]:
            y = (mt_vars["mt"][kk] & mt_vars["UPPER_MASK"]) | (
                mt_vars["mt"][kk + 1] & mt_vars["LOWER_MASK"]
            )  # y = (mt[kk] AND UPPER_MASK) OR (mt[kk + 1] AND LOWER_MASK)
            mt_vars["mt"][kk] = (
                mt_vars["mt"][kk + mt_vars["M"]] ^ (y >> 1) ^ mag01[y & 0x1]
            )  # mt[kk] = mt[kk + M] XOR (y >> 1) XOR mag01[y AND 0x1]
            kk += 1

        while kk < mt_vars["N"] - 1:
            y = (mt_vars["mt"][kk] & mt_vars["UPPER_MASK"]) | (
                mt_vars["mt"][kk + 1] & mt_vars["LOWER_MASK"]
            )  # y = (mt[kk] AND UPPER_MASK) OR (mt[kk + 1] AND LOWER_MASK)
            mt_vars["mt"][kk] = (
                mt_vars["mt"][kk + (mt_vars["M"] - mt_vars["N"])]
                ^ (y >> 1)
                ^ mag01[y & 0x1]
            )  # mt[kk] = mt[kk + (M - N)] XOR (y >> 1) AND mag01[y & 0x1]
            kk += 1

        y = (mt_vars["mt"][mt_vars["N"] - 1] & mt_vars["UPPER_MASK"]) | (
            mt_vars["mt"][0] & mt_vars["LOWER_MASK"]
        )  # y = (mt[N - 1] AND UPPER_MASK) OR (mt[0] AND LOWER_MASK)
        mt_vars["mt"][mt_vars["N"] - 1] = (
            mt_vars["mt"][mt_vars["M"] - 1] ^ (y >> 1) ^ mag01[y & 0x1]
        )  # mt[N - 1] = mt[M - 1] XOR (y >> 1) XOR mag01[y & 0x1]

        mt_vars["mti"] = 0

    y = mt_vars["mt"][mt_vars["mti"]]  # y = mt[mti++]
    mt_vars["mti"] += 1

    # Tempering
    y ^= mt_vars["TEMPERING_SHIFT_U"](y)
    y ^= mt_vars["TEMPERING_SHIFT_S"](y) & mt_vars["TEMPERING_MASK_B"]
    y ^= mt_vars["TEMPERING_SHIFT_T"](y) & mt_vars["TEMPERING_MASK_C"]
    y ^= mt_vars["TEMPERING_SHIFT_L"](y)

    return y / 0xFFFFFFFF



import numpy as np
from BNumMet.Interpolation import polinomial

global exceptions
exceptions = [ValueError("The function has no zeros in the given interval")]


def bisect(f, interval, stop_iters=100, iters=False, *args):
    """
    Finds a zeros over the given interval using the Bisection method and tolerance ~ machine precision

    params:
        f: function to find the zeros
        interval: interval where the zeros are searched
        *args: arguments of the function f

    returns:
        x: zeros of the function f

    raises:
        ValueError: if the function has no zeros in the given interval

    """

    x0, x1 = interval
    # Evaluate the function at the two points of the interval
    f0 = f(x0, *args)
    f1 = f(x1, *args)

    # If the function has the same sign at the two points, raise an error as there's no zero between them
    if f0 * f1 > 0:
        raise exceptions[0]

    # Initialize x to the right endpoint of the interval
    x = x1
    # Initialize the iteration count to 0
    iterations = 0
    # Repeat the following loop until the function value at x is zero or the iteration count exceeds the maximum
    while f1 and iterations < stop_iters:
        # Increment the iteration count
        iterations += 1
        # Update x to be the midpoint of the interval
        x = 0.5 * (x0 + x1)
        # Evaluate the function at the new x
        faux = f(x, *args)
        # If the function has different signs at x and x1, set x0 to x
        if faux * f1 < 0:
            x0 = x
        # If the function has the same sign at x and x1, set x1 to x and update f1
        else:
            x1 = x
            f1 = faux
    # If the iters argument is set to True, return the zero of the function and the number of iterations taken
    if iters:
        return x, iterations
    # If the iters argument is set to False (default), return just the zero of the function
    return x


def secant(fun, interval, stop_iters=100, iters=False, *args):
    """
    Finds a zeros over the given interval using the secant method

    params:
        fun: function to find the zeros
        interval: interval where the zeros are searched
        *args: arguments of the function fun
        stop_iters: maximum number of iterations (default 100)
        iters: a flag to return the number of iterations performed (default False)

    returns:
        x: zeros of the function fun

    raises:
        ValueError: if the function has no zeros in the given interval

    """
    # Set initial values for x0, x1
    x0, x1 = interval
    # Evaluate the function at x0 and x1
    f0 = fun(x0, *args)
    f1 = fun(x1, *args)

    # Check if the product of f0 and f1 is greater than 0
    # If yes, raise a ValueError, as there is no zero in the interval
    if f0 * f1 > 0:
        raise ValueError("The function has no zeros in the given interval")

    # Initialize the number of iterations
    iterations = 0
    # Use the secant method to find the zero
    while abs(x1 - x0) > np.finfo(float).eps and iterations < stop_iters:
        # Increase the iteration count
        iterations += 1
        # Update the values of x2, x0, and x1
        x2 = x0
        x0 = x1
        x1 = x1 + (x1 - x2) / (fun(x2, *args) / fun(x1, *args) - 1)

    # Check if the flag `iters` is set to True
    if iters:
        # Return the zero and the number of iterations
        return x1, iterations
    # Return the zero
    return x1


def newton(fun, derivative, start_point, stop_iters=100, iters=False, *args):
    """
    Finds a zeros over the given interval using the Newton-Raphson method

    params:
        f: function to find the zeros
        interval: interval where the zeros are searched
        *args: arguments of the function f

    returns:
        x: zeros of the function f

    raises:
        ValueError: if the function has no zeros in the given interval

    """
    # initializing previousX with a value not equal to startPoint
    # and xn with the startPoint to allow for iteration
    previous_x = start_point - 1
    xn = start_point

    # evaluating the function at xn
    fn = fun(xn, *args)

    # checking if the derivative of the function at xn is zero
    # and raising an error if it is
    if derivative(xn, *args) == 0:
        raise ValueError("The derivative of the function is zero")

    # initializing iteration count to zero
    iterations = 0

    # checking if the function at xn is not equal to zero
    # and that xn is not equal to previousX and the derivative
    # of the function at xn is not equal to zero and that
    # the number of iterations is less than the stopping criteria
    while (
        fn != 0
        and not np.isclose(xn - previous_x, 0)
        and derivative(xn, *args) != 0
        and iterations < stop_iters
    ):
        # incrementing iteration count by one
        iterations += 1

        # updating previousX with the current value of xn
        previous_x = xn
        # updating xn with the next iteration value
        # computed using the Newton-Raphson method
        xn = xn - fn / derivative(xn, *args)
        # evaluating the function at xn
        fn = fun(xn, *args)

    # if the iters flag is set, return xn and the number of iterations
    # otherwise return only xn
    if iters:
        return xn, iterations
    return xn


def IQI(f, x_values, stop_iters=100, iters=False, *args):
    """
    Finds a zeros over the given interval using the Inverse Quadratic Interpolation method

    params:
        f: function to find the zeros
        xVals : [x0,x1,x2]
        *args: arguments of the function f

    returns:
        x: zeros of the function f

    raises:
        ValueError: if the function has no zeros in the given interval

    """
    # Unpacking the initial values for x0, x1 and x2
    x0, x1, x2 = x_values
    # Initializing the iteration counter to 0
    iterations = 0
    # The main loop, where the inverse quadratic interpolation is performed
    # This loop continues until either the relative difference between x1 and x0 is less than the machine precision
    # or the number of iterations has reached the maximum allowed iterations
    while abs(x1 - x0) > np.finfo(float).eps and iterations < stop_iters:
        # Increase the iteration counter
        iterations += 1
        # Evaluate the function at x0, x1, and x2
        f0, f1, f2 = f(x0, *args), f(x1, *args), f(x2, *args)
        # Perform the inverse quadratic interpolation
        aux1 = (x0 * f1 * f2) / ((f0 - f1) * (f0 - f2))
        aux2 = (x1 * f0 * f2) / ((f1 - f0) * (f1 - f2))
        aux3 = (x2 * f1 * f0) / ((f2 - f0) * (f2 - f1))
        new = aux1 + aux2 + aux3
        # Shift x0, x1, x2 to the right by 1 position
        x0, x1, x2 = new, x0, x1

    # If the argument iters is set to True, return both the result and the number of iterations
    if iters:
        return x0, iterations
    # Otherwise, return only the result
    return x0


def zBrentDekker(
    f, interval, tol=10 ** (-20), stop_iters=100, iters=False, steps=False, *args
):
    """
    Finds a zeros over the given interval using a combination of Bisection and secant method

    params:
        f: function to find the zeros
        interval: interval where the zeros are searched
        *args: arguments of the function f

    returns:
        x: zeros of the function f

    raises:
        ValueError: if the function has no zeros in the given interval

    """
    # Split the interval into a and b
    a, b = interval
    # Evaluate the function at a and b
    fa = f(a, *args)
    fb = f(b, *args)

    # Check if there are no zeros in the interval
    if fa * fb > 0:  # No zeros guaranteed in the interval
        raise exceptions[0]

    # Initialize the variables for the internal section
    c, fc, d, e = a, fa, b - a, b - a

    # Check if fc is smaller than fb, if so swap the values
    if abs(fc) < abs(fb):
        a, b, c, fa, fb, fc = b, c, b, fb, fc, fb

    # Calculate the tolerance level
    tolerance = 2 * np.finfo(float).eps * abs(b) + tol
    m = 0.5 * (c - b)

    # Initialize iteration and procedure stack
    iterations = 0
    procedure_stack = []

    # Repeat until the tolerance level is met or max iterations is reached
    while abs(m) > tolerance and fb and iterations < stop_iters:
        # Calculate next step
        # =============================================================================================================
        # Check if bisection is forced
        if abs(e) < tolerance or abs(fa) <= abs(fb):
            d = m
            e = m
            procedure_stack.append("Bisection")
        else:
            # Calculate the ratio of fb and fa
            s = fb / fa
            if a == c:
                # Use linear interpolation
                p = 2 * m * s
                q = 1 - s
            else:
                # Use inverse quadratic interpolation
                q = fa / fc
                r = fb / fc
                p = s * (2 * m * q * (q - r) - (b - a) * (r - 1))
                q = (q - 1) * (r - 1) * (s - 1)

            # Correct the sign of p and q
            if p > 0:
                q = -q
            else:
                p = -p

            s = e
            e = d

            # Validate the interpolation
            if 2 * p < 3 * m * q - abs(tolerance * q) and p < abs(0.5 * s * q):
                # The interpolation is valid
                d = p / q
                procedure_stack.append("IQI" if a != c else "Secant")
            else:
                # The interpolation is not valid, we use bisection
                d = m
                e = m
                procedure_stack.append("Bisection")
        a = b
        fa = fb

        b += d if abs(d) > tolerance else np.sign(m) * tolerance
        fb = f(b, *args)
        # =============================================================================================================

        # Update interval
        # =============================================================================================================
        # Correct points accordingly
        if np.sign(fb) == np.sign(fc) != 0:
            # Section: int
            c, fc, d, e = a, fa, b - a, b - a
        # Section: ext
        elif abs(fc) < abs(fb):
            a, b, c, fa, fb, fc = b, c, b, fb, fc, fb

        # Update tolerance and m for the next iteration
        tolerance = 2 * np.finfo(float).eps * abs(b) + tol
        m = 0.5 * (c - b)

        iterations += 1
        # print(b,e)
    zero = b
    if steps and iters:
        return zero, iterations, procedure_stack
    if iters:
        return zero, iterations
    if steps:
        return zero, procedure_stack
    return zero



import sys
import os
from unittest import TestCase
import unittest
import numpy as np
from random import randint

from BNumMet.Visualizers.LeastSquaresVisualizer import LSPVisualizer as LSV


class Test_LSPVisualizer(TestCase):
    def test_init(self):
        # Initialize the object with no data
        lsp = LSV()
        self.assertTrue(
            all(lsp.x_data_lsp != None)
        )  # Check that the x_data_lsp is not None
        self.assertTrue(
            all(lsp.y_data_lsp != None)
        )  # Check that the y_data_lsp is not None

    def test_init_with_data(self):
        # Initialize the object with data
        x_data_lsp = np.array([1, 2, 3, 4, 5])
        y_data_lsp = np.array([1, 2, 3, 4, 5])
        lsp = LSV(x_data_lsp, y_data_lsp)
        self.assertEqual(np.array_equal(lsp.x_data_lsp, x_data_lsp), True)
        self.assertEqual(np.array_equal(lsp.y_data_lsp, y_data_lsp), True)

    def test_exceptionInit(self):
        # Initialize the object with data of different sizes
        x_data_lsp = np.array([1, 2, 3, 4, 5])
        y_data_lsp = np.array([1, 2, 3, 4])
        self.assertRaises(Exception, LSV, x_data_lsp, y_data_lsp)

        # len(x_data_lsp) < 2
        x_data_lsp = np.array([1])
        y_data_lsp = np.array([1])
        self.assertRaises(Exception, LSV, x_data_lsp, y_data_lsp)

    def test_run(self):
        # Run the object with no data
        lsp = LSV()
        # No exception should be raised
        lsp.run()

    def test_select_poly(self):
        # Run the object with no data
        lsp = LSV()
        lsp.run()

        lsp.function_type.value = "Polynomial"  # No Exception should be raised
        self.assertTrue(
            len(lsp.remarks.value) > 1
        )  # Check that the remarks are not empty))

        # Select the polynomial degree
        for i in range(lsp.polynomial_degree.max):
            lsp.polynomial_degree.value = i  # No Exception should be raised

        lsp.polynomial_degree.value = lsp.polynomial_degree.max + 1
        self.assertTrue(
            lsp.polynomial_degree.value == lsp.polynomial_degree.max
        )  # Check that the degree is not greater than the max

    def test_select_data(self):
        # Run the object with no data
        lsp = LSV()
        lsp.run()

        lsp.function_type.value = "Only data"
        self.assertTrue(len(lsp.remarks.value) == 0)  # Check that the remarks are empty

    def test_select_exp(self):
        # Run the object with no data
        lsp = LSV()
        lsp.run()

        lsp.function_type.value = "Exponential"  # No Exception should be raised
        self.assertTrue(
            len(lsp.remarks.value) > 1
        )  # Check that the remarks are not empty))

    def test_select_sinCos(self):
        # Run the object with no data
        lsp = LSV()
        lsp.run()

        lsp.function_type.value = "Sines & Cosines"
        self.assertTrue(
            len(lsp.remarks.value) > 1
        )  # Check that the remarks are not empty))

        # Select the Sines & Cosines degree
        for i in range(lsp.sine_cosine_degree.max):
            lsp.sine_cosine_degree.value = i

        lsp.sine_cosine_degree.value = lsp.sine_cosine_degree.max + 1
        self.assertTrue(
            lsp.sine_cosine_degree.value == lsp.sine_cosine_degree.max
        )  # Check that the degree is not greater than the max



import numpy as np
from ..Interpolation import polinomial, piecewise_linear, pchip, splines
from bqplot import pyplot as plt
import ipywidgets as widgets
import bqplot as bq


class InterpolVisualizer:
    def __init__(
        self,
        x_initial=list(np.arange(1, 7, 1).astype(float)),
        y_initial=[16, 18, 21, 17, 15, 12],
        u_initial=list(np.arange(1, 6.1, 0.1)),
    ) -> None:
        """
        Initializes the Class

        Parameters:
            - xInitial: Initial x coordinates
            - yInitial: Initial y coordinates
            - uInitial: Initial mesh
        """
        if len(x_initial) != len(y_initial):
            raise ValueError("The length of the X and Y coordinates must be the same")

        self.x = np.array(x_initial).astype(float)
        self.y = np.array(y_initial).astype(float)
        self.u = np.array(u_initial).astype(float)

        self.originals = [self.x, self.y, self.u]

        self.methods = {
            "InterPoly": [polinomial, "blue"],
            "Piecewise Linear": [piecewise_linear, "green"],
            "Pchip": [pchip, "orange"],
            "Splines": [splines, "purple"],
        }

    def initialize_components(self):
        """
        Initializes the components of the GUI
        Components:
            - Checkboxes
                Checkboxes for each interpolation method
            - Slider
                Slider for the mesh size
            - Reset Button
                Button to reset the points
        """

        """
        Checkboxes
        ======================
        One checkbox for every possible interpolation method
        Every checkbox will be associated with a method called update_checkboxes that will update the plot according to the updated checkboxes. 
        Additionally, the checkbox will have the same color as the color of associated interpolation method
        """
        self.checkboxes = []
        for key, val in self.methods.items():
            checkbox = widgets.Checkbox(
                description=key, value=True, style={"background-color": val[1]}
            )
            checkbox.background_color = val[1]
            checkbox.observe(self.update_checkboxes, "value")

            self.methods[key].append(checkbox)
            self.checkboxes.append(checkbox)

        """
        Reset button
        ======================
        Button to reset the points to the original ones, linked to the method reset
        """
        self.reset_button = widgets.Button(
            description="Reset", button_style="danger", tooltip="Reset", icon="undo"
        )
        self.reset_button.on_click(self.reset)

        """
        block adding points checkbox
        ======================
        Checkbox to block adding points, linked to anything - it is just a boolean value that will be used in update_X and update_Y
        """
        self.block_adding = widgets.Checkbox(
            description="Block adding points", value=False
        )

        """
        Effects of Extrapolation checkbox
        ======================
        Checkbox to show the effects of extrapolation, linked to the method update_extrapolation
        """
        self.extrapolation = widgets.Checkbox(
            description="Show effects of extrapolation", value=False
        )
        self.extrapolation.observe(self.update_extrapolation, "value")

        """
        Auto Zoom button
        ======================
        Button to auto zoom the plot, linked to the method auto_zoom
        """
        self.auto_zoom_button = widgets.Button(
            description="Auto Zoom",
            button_style="info",
            tooltip="Auto Zoom",
            icon="search-plus",
        )
        self.auto_zoom_button.on_click(self.auto_zoom)

    def update_extrapolation(self, change):
        """
        Updates the plot according to the effects of extrapolation checkbox
        """
        min_max = [min(self.x), max(self.x)]
        half = (min_max[1] - min_max[0]) / 4
        if change["new"]:
            self.u = np.linspace(
                min_max[0] - half, min_max[1] + half, 100, endpoint=True
            )
        else:
            self.u = np.linspace(min_max[0], min_max[1], 100, endpoint=True)

        self.update_x(None)
        self.update_y(None)

    def auto_zoom(self, change):
        """
        Auto zooms the plot
        """

        self.x_sc.min = min(self.u)
        self.x_sc.max = max(self.u)
        y_values = [j for line in self.interpolation_lines for j in line.y]
        if len(y_values) != 0:
            self.y_sc.min = min(y_values)
            self.y_sc.max = max(y_values)

    def scatter_dots(self):
        """
        Updates ScatteredDots - the scatter plot of the data - according to the new points
        It observers the changes in the x and y coordinates of the scatter plot and links them to the update_X and update_Y methods
        It also lets adding new points
        """
        self.scattered_dots = bq.Scatter(
            x=self.x,
            y=self.y,
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["red"],
            name="Points",
            enable_move=True,
            enable_add=False,
            display_legend=False,
            labels=["Points"],
        )

        # observe change XY
        self.scattered_dots.observe(self.update_x, "x")
        self.scattered_dots.observe(self.update_y, "y")
        self.scattered_dots.interactions = {"click": "add"}

    def interpol_lines(self):
        """
        Updates the interpolation lines according to the new points
        It creates an array of Lines for every interpolation method that is checked in the checkboxes
        """
        self.interpolation_lines = [
            bq.Lines(
                x=self.u,
                y=val[0](self.x, self.y, self.u),
                scales={"x": self.x_sc, "y": self.y_sc},
                colors=[val[1]],
                name=key,
                display_legend=False,
                labels=[key],
                enable_move=False,
                enable_add=False,
            )
            for key, val in self.methods.items()
            if val[2].value
        ]

    def update_x(self, change):
        """
        Updates the x coordinates and the plot according to the new x coordinates if the change is not None and does not contain Repetitions (Definition of a function)
        It also updates the slider according to the new x coordinates

        This method will always be called when the x coordinates are changed and before the y coordinates are changed.
        """
        if (
            change is not None
            and change["name"] == "x"
            and len(list(change["new"])) == len(set(change["new"]))
        ):  # There are changes and there are no repetitions in the x coordinates
            if (
                len(list(change["new"])) > len(self.x) and not self.block_adding.value
            ):  # There are more points than before and the block adding points checkbox is not checked
                self.x = change["new"]
            elif len(list(change["new"])) <= len(
                self.x
            ):  # There are the same number of points as before - the points are moved
                self.x = change["new"]
            else:  # There are more points than before and the block adding points checkbox is checked
                return  # Nothing to re-plot

    def update_y(self, change):
        """
        Updates the y coordinates and the plot according to the new y coordinates if the change is not None and contains the same number of points as X
        It makes new scattering and inteprolation lines and updates the plot.
        Also, it updates Scales and Axes according to the new points

        This method will always be called when the y coordinates are changed and after the x coordinates are changed.
        """

        with self.widgetsgrid.hold_sync():
            if (
                change is not None
                and change["name"] == "y"
                and len(list(change["new"])) == len(self.x)
            ):
                if (
                    len(list(change["new"])) > len(self.y)
                    and not self.block_adding.value
                ):
                    self.y = change["new"]
                elif len(list(change["new"])) <= len(self.y):
                    self.y = change["new"]
                else:
                    return  # Nothing to re-plot

            self.scatter_dots()
            self.interpol_lines()

            to_update = [*self.interpolation_lines, self.scattered_dots]

            self.fig.marks = to_update

    def update_checkboxes(self, change):
        """
        Updates the plot according to the new checkboxes
        """
        self.update_x(None)
        self.update_y(None)

    def reset(self, b):
        """
        Resets everything to what it was at the beginning of the program
        """
        self.x = self.originals[0]
        self.y = self.originals[1]
        self.u = self.originals[2]

        # Reset checkboxes
        for key, val in self.methods.items():
            val[2].value = True

        self.extrapolation.value = False
        self.block_adding.value = False

        values = [min(self.x), max(self.x)]
        self.u = np.linspace(values[0], values[1], 100, endpoint=True)

        self.update_x(None)
        self.update_y(None)

        # Reset Y scale
        self.y_sc.min = min(self.y)
        self.y_sc.max = max(self.y)

        # Reset X scale
        self.x_sc.min = min(self.x)
        self.x_sc.max = max(self.x)

    def run(self):
        """
        Runner method : Creates all the widgets and displays the plot with a given layout
        """
        self.x_sc = bq.LinearScale(stabilize=True)
        self.y_sc = bq.LinearScale(stabilize=True)
        ax_x = bq.Axis(scale=self.x_sc, grid_lines="solid", label="X")
        ax_y = bq.Axis(
            scale=self.y_sc,
            orientation="vertical",
            tick_format="0.2f",
            grid_lines="solid",
            label="Y",
        )

        self.initialize_components()
        self.interpol_lines()
        self.scatter_dots()

        # Reset Y scale
        self.y_sc.min = min(self.y)
        self.y_sc.max = max(self.y)

        # Reset X scale
        self.x_sc.min = min(self.x)
        self.x_sc.max = max(self.x)

        self.fig = bq.Figure(
            marks=[*self.interpolation_lines, self.scattered_dots],
            axes=[ax_x, ax_y],
            title="Interpolation Visualizer",
            legend_location="top-right",
            animation_duration=1000,
        )

        self.toolbar = bq.Toolbar(figure=self.fig)

        self.checkboxes_vbox = widgets.VBox(
            [
                widgets.HBox(
                    [
                        widgets.Button(
                            description="",
                            style={"button_color": i.background_color},
                            disabled=True,
                            layout=widgets.Layout(width="20px"),
                        ),
                        i,
                    ]
                )
                for i in self.checkboxes
            ]  # BUTTON + CHECKBOX PAIR -- Fake Legend
        )
        tools = widgets.VBox(
            [
                self.checkboxes_vbox,
                self.block_adding,
                self.extrapolation,
                widgets.HBox([self.reset_button, self.auto_zoom_button]),
            ]
        )

        self.widgetsgrid = widgets.GridspecLayout(11, 7)

        self.widgetsgrid[0, :] = self.toolbar
        self.widgetsgrid[1:, :4] = self.fig
        self.widgetsgrid[1:, 4:] = tools

        return self.widgetsgrid



import numpy as np
from bqplot import pyplot as plt
import ipywidgets as widgets
import bqplot as bq
from ..LinearSystems import qr_solve


class LSPVisualizer:
    """
    Visualizer for the least squares method with the examples of
        * Exponential function
        * Polynomial function
        * Sine and Cosine functions
    """

    def __init__(
        self, x_data_lsp=np.array([0, 1, 2, 3]), y_data_lsp=np.array([4.5, 2.4, 1.5, 1])
    ):
        if len(x_data_lsp) != len(y_data_lsp):  # check if the data is valid
            raise ValueError("x_data_lsp and y_data_lsp must have the same length")
        if len(x_data_lsp) < 2:  # check if the data is valid
            raise ValueError(
                "x_data_lsp and y_data_lsp must have at least 2 points... Too boring otherwise"
            )
        self.x_data_lsp = np.array(
            x_data_lsp, dtype=np.float64
        )  # convert the data to float64 (float64 is the default type for numpy)
        self.y_data_lsp = np.array(
            y_data_lsp, dtype=np.float64
        )  # convert the data to float64 (float64 is the default type for numpy)

        min_max_xdata = np.min(self.x_data_lsp), np.max(
            self.x_data_lsp
        )  # min and max of x_data_lsp
        self.x = np.linspace(  # x values for plotting the fitted curve
            np.min(x_data_lsp) - 0.2 * (min_max_xdata[1] - min_max_xdata[0]),  # min x
            np.max(x_data_lsp) + 0.2 * (min_max_xdata[1] - min_max_xdata[0]),  # max x
            1000,  # number of points
        )  # for plotting the fitted curve

    def initialize_components(self):
        """
        Initialize the components of the visualizer
        """

        # 1. Dropdown for selecting the function type
        self.function_type = widgets.Dropdown(  #
            options=[
                "Only data",
                "Polynomial",
                "Exponential",
                "Sines & Cosines",
            ],  # options
            value="Only data",  # default value
            description="Function:",  # description
            disabled=False,  # disable the dropdown
        )
        self.function_type.observe(
            self.selector_function, names="value"
        )  # observe the change of the dropdown
        # 2. Int text for selecting the polynomial degree
        self.polynomial_degree = widgets.BoundedIntText(  # int text for selecting the polynomial degree
            value=1,  # default value
            description="Degree:",  # description
            disabled=False,
            # make smaller width
            layout=widgets.Layout(width="75%"),
            min=0,  # minimum value
        )
        self.polynomial_degree.observe(
            self.degree_change_poly, names="value"
        )  # observe the change of the int text
        # 3. Int text for selecting the number of sines and cosines
        self.sine_cosine_degree = widgets.BoundedIntText(  # int text for selecting the number of sines and cosines
            value=1,  # default value
            description="",  # description
            disabled=False,
            min=0,
            layout=widgets.Layout(width="50%")
            # Make smaller width
        )

        self.sine_cosine_degree.observe(
            self.degree_change_sine_cosine, names="value"
        )  # observe the change of the int text

        self.sine_cosine_degree_box = widgets.HBox(
            [widgets.Label("Basis\n Elements:"), self.sine_cosine_degree]
        )

        # 4. Checkbox for Error Bound Visualization
        self.error_bound = widgets.Checkbox(  # checkbox for error bound visualization
            value=False,  # default value
            description="Error Bound",  # description
            disabled=False,  # disable the checkbox
        )

        # 5. Remarks about LSP output Math
        self.remarks = widgets.HTMLMath(  # remarks about LSP output
            value="",  # default value
            placeholder="LSP Remarks",  # placeholder
            description="Remarks:",  # description
        )

        # 6. Figure
        self.x_sc = bq.LinearScale()  # x scale
        self.y_sc = bq.LinearScale()  # y scale
        ax_x = bq.Axis(scale=self.x_sc, grid_lines="solid", label="X")
        ax_y = bq.Axis(
            scale=self.y_sc,
            orientation="vertical",
            tick_format="0.4e",
            grid_lines="solid",
            label="Y",
        )
        # Set up the plot figure
        self.fig = bq.Figure(
            marks=[],  # marks
            axes=[ax_x, ax_y],  # axes
            title="Least Squares Method",  # title
        )

        self.grid = widgets.GridspecLayout(4, 4)  # grid for the components
        self.grid[:, 0:2] = self.fig  # add the figure to the grid
        self.grid[0:2, 2:] = widgets.VBox(
            [self.function_type]
        )  # add the function type dropdown to the grid
        # self.grid[1, 3] --> self.errorBound
        self.grid[2:, 2:] = self.remarks  # add the remarks to the grid

        # Plot the data (Since it will be the same for all the functions)
        points = bq.Scatter(  # data points
            x=self.x_data_lsp,  # x values
            y=self.y_data_lsp,  # y values
            scales={"x": self.x_sc, "y": self.y_sc},  # scales
        )
        self.fig.marks = [points]  # add the data points to the figure

    def degree_change_poly(self, change):
        """
        This function is called when the polynomial degree is changed
        """
        curve, c, err = self.polynomial_lsp(change["new"])  # calculate the curve
        self.fig.marks = [self.fig.marks[0], curve]  # update the figure
        self.polynomial_remarks(c, err)  # update the remarks

    def degree_change_sine_cosine(self, change):
        """
        This function is called when the sine and cosine degree is changed
        """
        curve, c, err = self.sine_cosine_lsp(change["new"])  # calculate the curve
        self.fig.marks = [self.fig.marks[0], curve]  # update the figure
        self.sine_cosine_remarks(c, err)  # update the remarks

    def selector_function(self, change):
        """
        This function is called when the function type is changed
        """
        to_draw = [self.fig.marks[0]]  # the data points are always drawn
        to_grid = [self.function_type]  # the function type dropdown
        if (
            change["new"] == "Only data"
        ):  # if the function type is "Only data"  then we don't need to draw anything else
            self.remarks.value = ""  # clear the remarks
        elif (
            change["new"] == "Polynomial"
        ):  # if the function type is "Polynomial" then we need to
            to_grid.append(
                self.polynomial_degree
            )  # add the polynomial degree dropdown to the grid
            self.polynomial_degree.max = (  # set the max degree to the number of data points - 1
                len(self.x_data_lsp) - 1
            )  # max degree is number of data points - 1
            curve, c, err = self.polynomial_lsp(
                self.polynomial_degree.value
            )  # calculate the polynomial LSP
            to_draw.append(curve)  # add the polynomial to the list of marks to draw
            self.polynomial_remarks(c, err)  # add the remarks to the remarks widget

        elif (
            change["new"] == "Exponential"
        ):  # if the function type is "Exponential" then we need to
            curve, c, err = self.exponential_lsp()  # calculate the exponential LSP
            to_draw.append(curve)  # add the exponential to the list of marks to draw
            self.exponential_remarks(c, err)  # add the remarks to the remarks widget

        elif (
            change["new"] == "Sines & Cosines"
        ):  # if the function type is "Sines & Cosines" then we need to
            to_grid.append(
                self.sine_cosine_degree_box
            )  # add the sine and cosine degree dropdown to the grid
            self.sine_cosine_degree.max = (
                (  # set the max degree to half the number of data points - 1
                    len(self.x_data_lsp) - 1
                )
                // 2
            )  # max degree is half the number of data points - 1
            curve, c, err = self.sine_cosine_lsp(
                self.sine_cosine_degree.value
            )  # calculate the sine and cosine LSP
            to_draw.append(
                curve
            )  # add the sine and cosine to the list of marks to draw
            self.sine_cosine_remarks(c, err)  # add the remarks to the remarks widget

        with self.fig.hold_sync():  # update the figure
            self.fig.marks = to_draw  # update the marks
            self.grid[0:2, 2:] = widgets.VBox(
                to_grid,
                layout=widgets.Layout(  # set the layout of the grid
                    width="auto",
                    height="auto",
                ),
            )  # update the grid (the dropdowns and degree selectors)

    def exponential_lsp(self):
        """
        This function returns the exponential function that best fits the data Using LSP

        Returns
        -------
        curve : bqplot.Lines
            The curve that best fits the data
        a : float
            The coefficient of the exponential function
        b : float
            The coefficient of the exponential function
        """
        # Create the A matrix
        A = np.ones((len(self.x_data_lsp), 2))  # 2 columns for the 2 coefficients
        A[:, 1] = self.x_data_lsp  # The second column is the x data
        # Solve the system
        c = qr_solve(
            A, np.log(self.y_data_lsp)
        )  # Solve the system using QR decomposition

        # Revert the log on c[0]
        c[0] = np.exp(c[0])  # We have made the log of c[0] to solve the system

        # Plot the curve
        # Evaluate the exponential function
        y = c[0] * np.exp(c[1] * self.x)  # Evaluate the exponential function
        # Plot the curve
        curve = bq.Lines(
            x=self.x,
            y=y,
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["red"],
        )  # Create the curve object

        err = np.linalg.norm(
            np.log(self.y_data_lsp) - A @ np.array([np.log(c[0]), c[1]]), 2
        )  # Calculate the error

        return curve, c, err

    def exponential_remarks(self, c, err):
        """
        This function updates the remarks section of the widget

        Remark: The exponential function that best fits the data is:
            <br>$y = a e^{bx}$ <br><br>
        The coefficients of the exponential function are:
            <br>$a = $ <br>$b = $ <br><br>
        The error is:
            $err$

        Parameters
        ----------
        c : np.array
            The coefficients of the exponential function in the form of c[0]*x^c[1]
        """
        self.remarks.value = f"The exponential function that best fits the data is: <br>$y = {c[0]:.4f}e^{{{c[1]:.4f}x}}$ <br><br> The coefficients of the exponential function are: <br>$c_0 = {c[0]:.4f}$ <br>$c_1 = {c[1]:.4f}$ <br><br> The error is: ${err:.4f}$ $\left(\sqrt{{\sum_i (y_i - f(x_i))^2}}\\right)$"

    def polynomial_lsp(self, degree):
        """
        This function returns the polynomial function of the given degree that best fits the data Using LSP

        Parameters
        ----------
        degree : int
            The degree of the polynomial function

        Returns
        -------
        curve : bqplot.Lines
            The curve that best fits the data
        c : np.array
            The coefficients of the polynomial function in the form of c[0] + c[1]*x + c[2]*x^2 + ... + c[degree]*x^degree
        """
        # Create the A matrix
        A = np.zeros((len(self.x_data_lsp), degree + 1))  # +1 because of the 0th degree
        for i in range(degree + 1):  # +1 because of the 0th degree
            A[:, i] = (
                self.x_data_lsp**i
            )  # x^0 = 1 so no need to do anything special for the 0th degree case but everything else is x^i
        # Solve the system
        c = qr_solve(A, self.y_data_lsp)
        # Plot the curve
        # Evaluate the polynomial function
        y = np.array(
            [np.sum([c[i] * x**i for i in range(degree + 1)]) for x in self.x]
        )  # sum_{i=0}^n c_i x^i = c_0 + c_1 x + c_2 x^2 + ... + c_n x^n =Evaluate the polynomial function
        # Plot the curve
        curve = bq.Lines(
            x=self.x,
            y=y,
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["red"],
        )  # Create the curve object

        err = np.linalg.norm(self.y_data_lsp - A @ c)  # Calculate the error

        return curve, c, err  # Return the curve, coefficients and error

    def polynomial_remarks(self, c, err):
        """
        This function returns the remarks for the polynomial function

        The polynomial function is: <br> $y = c_0 + c_1x + c_2x^2 + ... + c_nx^n$ <br><br>
        The coefficients of the polynomial function are: <br> $c_0 = ...$ <br> $c_1 = ...$ <br> $c_2 = ...$ <br> ... <br> $c_n = ...$ <br><br>
        The error is: ...$"

        Parameters
        ----------
        c : np.array
            The coefficients of the polynomial function in the form of c[0] + c[1]*x + c[2]*x^2 + ... + c[degree]*x^degree

        Returns
        -------
        remarks : str
            The remarks for the polynomial function
        """

        # Create the remarks
        remarks = "The polynomial function is:<br> $"  # First line of the remarks

        for i in range(len(c)):  # Add the coefficients of the polynomial function
            if i == 0:
                remarks += f"{c[i]:.2f}"  # If the coefficient is 0, don't add x^{i}
            elif i == 1:
                remarks += ("+" if np.sign(c[i]) >= 0 else "") + f"{c[i]:.2f}x"
                # If the coefficient is 1, don't add power
            else:
                remarks += ("+" if np.sign(c[i]) >= 0 else "") + f"{c[i]:.2f}x^{i}"
                # If the coefficient is 1, add power

        remarks += "$ <br><br> The coefficients of the polynomial function are: <br> "  # Second line of the remarks

        for i in range(len(c)):  # Add the coefficients of the polynomial function
            remarks += f"$c_{i} = {c[i]:.2f}$<br> "  # Add the coefficient

        remarks += f"<br> The error is ${err:.4f}$ $\left(\sqrt{{\sum_i (y_i - f(x_i))^2}}\\right)$ <br><br>"  # Third line of the remarks

        self.remarks.value = remarks  # Update the remarks

    def sine_cosine_lsp(self, degree):
        """
        This function returns the sine and cosine function of the given degree that best fits the data Using LSP

        Parameters
        ----------
        degree : int
            The degree of the sine and cosine function

        Returns
        -------
        curve : bqplot.Lines
            The curve that best fits the data
        c : np.array
            The coefficients of the sine and cosine function in the form of c[0] + c[1]*sin(x) + c[2]*cos(x) + ... + c[degree]*sin(degree*x) + c[degree+1]*cos(degree*x)
        """
        # Create the A matrix
        A = np.ones(
            (len(self.x_data_lsp), 2 * degree + 1)
        )  # 2*degree+1 because we have degree sin and degree cos
        for i in range(
            1, degree + 1
        ):  # We start from 1 because we already have the first column of ones
            A[:, 2 * i - 1] = np.sin(i * self.x_data_lsp)  # 2*i-1 for the sin
            A[:, 2 * i] = np.cos(i * self.x_data_lsp)  # 2*i for the cos
        # Solve the system
        c = qr_solve(A, self.y_data_lsp)
        # Plot the curve
        # Evaluate the sine and cosine function
        y = np.array(
            [
                np.sum(
                    [
                        c[i] * np.sin(i * x) + c[i + 1] * np.cos(i * x)
                        for i in range(1, degree + 1)
                    ]
                )  # Find the sum of the sin and cos terms sum_{i=1}^{degree} c[2*i-1]*sin(i*x) + c[2*i]*cos(i*x)
                + c[0]  # The first coefficient is the constant
                for x in self.x  # For each x in our mesh
            ]
        )
        # Plot the curve
        curve = bq.Lines(
            x=self.x,
            y=y,
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["red"],
        )  # Make the curve object

        err = np.linalg.norm(
            self.y_data_lsp
            - [
                np.sum(
                    [
                        c[i] * np.sin(i * x) + c[i + 1] * np.cos(i * x)
                        for i in range(1, degree + 1)
                    ]
                )
                + c[0]
                for x in self.x_data_lsp
            ]
        )  # Calculate the error

        return curve, c, err

    def sine_cosine_remarks(self, c, err):
        """
        This function returns the remarks for the sine and cosine function

        Parameters
        ----------
        c : np.array
            The coefficients of the sine and cosine function in the form of c[0] + c[1]*sin(x) + c[2]*cos(x) + ... + c[degree]*sin(degree*x) + c[degree+1]*cos(degree*x)

        Returns
        -------
        remarks : str
            The remarks for the sine and cosine function
        """

        # Create the remarks
        remarks = "The sine and cosine function is:<br> $"
        # first coefficient
        remarks += f"{c[0]:.2f}"

        for i in range(1, len(c) // 2 + 1):
            remarks += (
                "+" if np.sign(c[2 * i - 1]) >= 0 else ""
            )  # sign of the coefficient of the sine function
            remarks += (
                f"{c[2*i - 1]:.2f}\sin({i}\cdot x)"  # coefficient of the sine function
            )
            remarks += (
                "+" if np.sign(c[2 * i]) >= 0 else ""
            )  # sign of the coefficient of the cosine function
            remarks += (
                f"{c[2*i]:.2f}\cos({i}\cdot x)"  # coefficient of the cosine function
            )

        remarks += "$ <br><br> The coefficients of the sine and cosine function are: <br> "  # Second part of the remarks (coefficients)

        for i in range(len(c)):  # Iterate through coefficients
            remarks += f"$c_{i} = {c[i]:.2f}$ <br>"  # coefficient

        remarks += f"<br> The error is: ${err:.4f}$ $\left(\sqrt{{\sum_i (y_i - f(x_i))^2}}\\right)$ <br><br>"  # Third part of the remarks (error)

        self.remarks.value = remarks  # Set the remarks

    def run(self):
        self.initialize_components()

        return self.grid



import bqplot as bq
import numpy as np
from ipywidgets import widgets
from BNumMet.NonLinear import zBrentDekker


class NonLinearVisualizer:
    def __init__(
        self, fun=lambda x: (x - 1) * (x - 4) * np.exp(-x), interval=(0, 3), tol=1e-3
    ):
        # Initialize basic Parameters
        self.brent_dekker = zBrentDekker(
            fun, interval, tol, iters=True, steps=True
        )  # Get the output of Brent Dekkers Alg --> (x, iters)

        # Set the input function, the interval for the function evaluation,
        # and the tolerance for the algorithm
        self.f = fun
        self.a, self.b = interval
        self.fa, self.fb = self.f(self.a), self.f(self.b)

        # Check that the function has a zero in the interval
        if self.f(self.a) * self.f(self.b) > 0:
            raise ValueError("The function has no zeros in the given interval")

        # Set the tolerance and initial values for data variables
        self.t = tol
        self.original_data = (self.a, self.b)
        self.iterations = 0

        # First Step of the Algorithm
        # ==========================================================================
        ## Section: INT
        # Set the initial values for internal variables
        self.c, self.fc, self.e = self.a, self.fa, self.b - self.a
        ## Section: EXT
        # Call the sectionEXT() function to perform the first external section of the algorithm
        self.section_ext()

        # Current Step Save Values
        # ==========================================================================
        # Save the current state of the algorithm in a tuple
        self.current_step = (self.a, self.b, self.c, self.e)

        # Revert Stack
        # ==========================================================================
        # Create an empty stack to keep track of previous states of the algorithm
        self.revert_stack = []

        # Draw mesh
        # ==========================================================================
        # Create a numpy array of 1000 evenly spaced values between the minimum and maximum
        # values of the function interval to use for plotting the function
        self.x = np.linspace(min(self.a, self.b), max(self.a, self.b), 1000)
        self.widen = False

    def checkbox_changed(self, change):
        with self.fig.hold_sync():  # Hold the figure syncronized
            self.draw_figures()  # Draw the figures

    def section_int(self):
        # Check if the function evaluates with the same sign on both sides of c
        if np.sign(self.fb) == np.sign(self.fc) != 0:
            # If the condition is met, update the values of c, fc and e
            self.c, self.fc, self.e = self.a, self.fa, self.b - self.a

    def section_ext(self):
        # Check which of the endpoints has the smaller absolute value of f
        if abs(self.fc) < abs(self.fb):
            # If the condition is met, swap the values of a, b and c with b, c and b respectively,
            # and update the values of fa, fb and fc accordingly
            self.a, self.b, self.c, self.fa, self.fb, self.fc = (
                self.b,
                self.c,
                self.b,
                self.fb,
                self.fc,
                self.fb,
            )

    def initialize_components(self):
        # Current Solution Text
        # ==========================================================================
        # Widget for displaying current solution output
        # Current Solution: (b, f(b))
        # Iterations: N
        self.current_solution_output = widgets.Output()

        # Helper Text
        # ==========================================================================
        # Widget for displaying helper output
        # Suggestion for next step: <Bisect/IQI/Secant>
        self.helper_output = widgets.Output()  # Next Possible Step: <Bisect/IQI/None>

        # Brent-Dekker Solution
        # ==========================================================================
        # Widget for displaying Brent-Dekker solution output
        # Brent-Dekker Solution: (x^, f(x^)) in N^ iterations
        self.brent_dekker_output = widgets.HTML(
            value=f"<blockquote> Brent-Dekker Solution: <b>({self.brent_dekker[0]:.4e}, {self.f(self.brent_dekker[0]):.4e})</b> in <b>{self.brent_dekker[1]}</b> iterations"
        )

        # Reset Button
        # ==========================================================================
        # Widget for reset button
        self.reset_button = widgets.Button(
            description="Reset",
            disabled=False,
            button_style="danger",
            tooltip="Reset",
            icon="undo",
        )
        self.reset_button.on_click(self.reset)

        # Revert Button
        # ==========================================================================
        # Widget for revert button
        self.revert_button = widgets.Button(
            description="Revert",
            disable=False,
            button_style="warning",
            tooltip="Revert",
            icon="arrow-left",
        )
        self.revert_button.on_click(self.revert)

        # FIGURE
        # ==========================================================================
        # Widget for plotting the function and its zeros
        # Set up axes and scales for the plot
        self.x_sc = bq.LinearScale()
        self.y_sc = bq.LinearScale()
        ax_x = bq.Axis(scale=self.x_sc, grid_lines="solid", label="X")
        ax_y = bq.Axis(
            scale=self.y_sc,
            orientation="vertical",
            tick_format="0.4e",
            grid_lines="solid",
            label="Y",
        )
        # Set up the plot figure
        self.fig = bq.Figure(
            marks=[],
            axes=[ax_x, ax_y],
            title="Zeros of a Function",
        )
        # Set up the plot figure toolbar
        self.toolbar = bq.Toolbar(figure=self.fig)
        # Add default lines to the plot figure
        self.default_lines()

        # FUNCTION BUTTONS
        # ==========================================================================
        # Widgets for buttons for selecting the next function to perform
        # Each button has a pointIndex attribute that is used to identify the point in the points array
        # 0: Bisect  1: Secant  2: IQI
        # Bisect Button
        self.bisect_button = widgets.Button(
            description="Bisect",
            disabled=False,
            tooltip="Bisect",
        )
        self.bisect_button.pointIndex = 0
        self.bisect_button.on_click(self.next_step)

        # Secant Button
        self.secant_button = widgets.Button(
            description="Secant",
            disabled=False,
            tooltip="Secant",
        )
        self.secant_button.pointIndex = 1
        self.secant_button.on_click(self.next_step)

        # IQI Button
        self.iqi_button = widgets.Button(
            description="IQI",
            disabled=False,
            tooltip="IQI",
        )
        self.iqi_button.pointIndex = 2
        self.iqi_button.on_click(self.next_step)

        # CHECKBOXES FOR FUNCTION BUTTONS
        # ==========================================================================
        self.bisect_checkbox = widgets.Checkbox(
            value=False,
            description="Bisect",
            disabled=False,
            indent=False,
        )
        # On change of checkbox, function checkBoxChanged
        self.bisect_checkbox.observe(self.checkbox_changed, names="value")

        self.secant_checkbox = widgets.Checkbox(
            value=False,
            description="Secant",
            disabled=False,
            indent=False,
        )
        self.secant_checkbox.observe(self.checkbox_changed, names="value")

        self.iqi_checkbox = widgets.Checkbox(
            value=False,
            description="IQI",
            disabled=False,
            indent=False,
        )
        self.iqi_checkbox.observe(self.checkbox_changed, names="value")

        # GRID
        # ==========================================================================
        self.grid = widgets.GridspecLayout(2, 2)
        self.grid[0, 0] = widgets.VBox([self.toolbar, self.fig])

        # Text Outputs Group
        texts_group = widgets.VBox(
            [self.current_solution_output, self.helper_output, self.brent_dekker_output]
        )
        self.grid[1, 0] = texts_group

        # Buttons Group
        text1 = widgets.HTML(value="<b>Next Step Selector</b>")  # Next Step Selector
        text2 = widgets.HTML(value="<b>Draw Step?</b>")  # Draw Step?
        selectors = widgets.VBox(  # Next Step Selectors
            [text1, self.bisect_button, self.secant_button, self.iqi_button]
        )
        checkboxes = widgets.VBox(
            [text2, self.bisect_checkbox, self.secant_checkbox, self.iqi_checkbox]
        )
        buttons_group = widgets.HBox([selectors, checkboxes])

        self.grid[0, 1] = buttons_group

        # Reset and Revert Buttons Group
        buttons_group_2 = widgets.HBox(
            [
                self.revert_button,
                self.reset_button,
            ]
        )
        self.grid[1, 1] = buttons_group_2

    def default_lines(self):
        # 0. Horizontal Line f(x)=0
        self.horizontal_line = bq.Lines(  # Horizontal Line f(x)=0
            x=self.x,
            y=[0] * len(self.x),
            scales={"x": self.x_sc, "y": self.y_sc},
            enable_move=False,
            enable_add=False,
            colors=["Black"],
        )
        # 1. Function Line f(x)
        self.function_line = bq.Lines(  # Function Line f(x)
            x=self.x,
            y=list(map(self.f, self.x)),
            scales={"x": self.x_sc, "y": self.y_sc},
            labels=["f(x)"],
            display_legend=False,
            enable_move=False,
            enable_add=False,
            colors=["Gray"],
            line_style="dashed",
        )

    def next_step(self, b):
        """
        This function is called when a button is clicked

        Parameters
        ----------
        b : Button
            The button that was clicked

        Returns
        -------
        None.
        """

        if self.hint_step is None:
            return
        self.revert_stack.append(
            [self.a, self.b, self.c, self.e, self.iterations]
        )  # Add the current state to the revert stack

        self.a = self.b  # Set a = b
        self.fa = self.fb  # Set f(a) = f(b)

        new_point_b = self.next_points_addition[b.pointIndex]  # Get the next point
        self.b = (
            new_point_b
            if abs(self.b - new_point_b) > self.tolerance
            else self.b + np.sign(0.5 * (self.c - self.b)) * self.tolerance
        )  # If the new point is too close to the current point, move it a little bit
        self.fb = self.f(self.b)  # Set f(b) = f(newB)
        if b.pointIndex != 1:
            self.e = self.errs[b.pointIndex]  # Set e = error
        self.iterations += 1  # Increment the number of iterations

        # Section: Ext
        self.section_ext()  # Update the section
        # Section: Int
        self.section_int()  # Update the section

        self.one_step()  # Update the plot

    def reset(self, b):
        """
        Reset everything to the initial state, this can be understood as reverting all the steps
        """
        if len(self.revert_stack) == 0:
            return

        self.a, self.b, self.c, self.e, self.iterations = self.revert_stack[
            0
        ]  # Get the initial state
        self.fb = self.f(self.b)  # Set f(b) = f(newB)
        self.fa = self.f(self.a)  # Set f(a) = f(b)
        self.fc = self.f(self.c)  # Set f(c) = f(newC)

        self.revert_stack = []  # Clear the revert stack
        self.one_step()  # Update the plot

    def revert(self, b):
        """
        This method reverts the last step
        """
        if len(self.revert_stack) == 0:  # If there is no step to revert
            return

        (
            self.a,
            self.b,
            self.c,
            self.e,
            self.iterations,
        ) = self.revert_stack.pop()  # Get the last state
        self.fb = self.f(self.b)  # Set f(b) = f(newB)
        self.fa = self.f(self.a)  # Set f(a) = f(b)
        self.fc = self.f(self.c)  # Set f(c) = f(newC)

        self.one_step()  # Update the plot

    def one_step(self):
        """
        This method is called when a step is made
        """
        with self.grid.hold_sync():  # Hold the grid syncronized
            self.brent_dekker_step()  # Update the Brent-Dekker step
            self.iqi_button.disabled = (
                self.next_points_addition[2] is None
            )  # Disable the IQI button if the IQI point is None
            self.iqi_checkbox.disabled = (
                self.next_points_addition[2] is None
            )  # Disable the IQI checkbox if the IQI point is None

            self.update_ouputs()  # Update the outputs
            with self.fig.hold_sync():  # Hold the figure syncronized
                self.draw_figures()  # Draw the figures
            # self.drawFigures()

    def draw_figures(self):
        """
        This method draws the figures
        1. The function
        2. Dot for the current solution
        3. The next step suggestions
        4. if checkboxes are checked, the steps are drawn
        """

        def draw_point(point, value, color, label, marker, legend=True):
            return bq.Scatter(
                x=[point],
                y=[value],
                scales={"x": self.x_sc, "y": self.y_sc},
                colors=[color],
                default_size=100,
                stroke="black",
                display_legend=legend,
                labels=[label],
                marker=marker,
            )  # Draw a point

        def secant_draw():
            # Draws a line on whole self.x range with the secant equation on the points (b, fb) and (self.nextPoints_addition[1], self.f(self.nextPoints_addition[1]))
            secant = bq.Lines(
                x=self.x,
                y=[
                    self.fb + (self.fa - self.fb) / (self.a - self.b) * (x - self.b)
                    for x in self.x
                ],
                scales={"x": self.x_sc, "y": self.y_sc},
                colors=["red"],
                display_legend=True,
                labels=["Secant Line"],
                line_style="dashed",
            )  # Draw a line
            return secant

        def bisect_draw(m, min_y, max_y):
            # m = (a + b) / 2
            # draw a vertical line at m
            bisect = bq.Lines(
                x=[m, m],
                y=[min_y, max_y],
                scales={"x": self.x_sc, "y": self.y_sc},
                colors=["green"],
                display_legend=True,
                labels=["Bisect Line"],
                line_style="dashed",
            )  # Draw a line
            return bisect

        def iqi_draw(a, b, c, min_y, max_y):
            # draw a vertical line at m
            inperpolation_y = [self.f(a), self.f(b), self.f(c)]  # [f(a), f(b), f(c)]
            interpolation_x = [a, b, c]  # [a, b, c]
            y_mesh = np.linspace(min_y, max_y, 1000)
            # Lagrange interpolation with (Y,X)
            q_y = (  # Lagrange interpolation
                lambda y: ((y - inperpolation_y[1]) * (y - inperpolation_y[2]))
                / (
                    (inperpolation_y[0] - inperpolation_y[1])
                    * (inperpolation_y[0] - inperpolation_y[2])
                )
                * interpolation_x[0]
                + ((y - inperpolation_y[0]) * (y - inperpolation_y[2]))
                / (
                    (inperpolation_y[1] - inperpolation_y[0])
                    * (inperpolation_y[1] - inperpolation_y[2])
                )
                * interpolation_x[1]
                + ((y - inperpolation_y[0]) * (y - inperpolation_y[1]))
                / (
                    (inperpolation_y[2] - inperpolation_y[0])
                    * (inperpolation_y[2] - inperpolation_y[1])
                )
                * interpolation_x[2]
            )
            x_mesh = q_y(y_mesh)  # Get the x values for the y values
            x_mesh = np.where(
                (x_mesh < self.x[0]) | (x_mesh > self.x[-1]), np.nan, x_mesh
            )  # Set the x values outside the range to nan
            iqi_line = bq.Lines(  # Draw a line
                x=x_mesh,
                y=y_mesh,
                scales={"x": self.x_sc, "y": self.y_sc},
                colors=["blue"],
                display_legend=True,
                labels=["IQI Line"],
                line_style="dashed",
            )
            return iqi_line

        points2check = [  # Points to check
            self.a,
            self.b,
            self.c,
            self.next_points_addition[0],
            self.next_points_addition[1],
        ] + (
            []
            if self.next_points_addition[2] is None
            else [
                self.next_points_addition[2]
            ]  # If the IQI point is None, add it to the list
        )

        if self.hint_step is None:  # If the hint step is None
            # set the original function
            self.x = np.linspace(min(self.original_data), max(self.original_data), 1000)
            self.default_lines()
            # Adjust the scales
            self.x_sc.min = min(self.original_data)
            self.x_sc.max = max(self.original_data)
            self.y_sc.min = min(self.function_line.y)
            self.y_sc.max = max(self.function_line.y)

            self.fig.marks = [
                self.horizontal_line,
                self.function_line,
                draw_point(  # Draw the current solution
                    self.b,
                    0,
                    "red",
                    "Current Solution",
                    "circle",
                ),
            ]  # Set the marks to the function and the horizontal line
            return

        min_max = [
            min(points2check),
            max(points2check),
        ]  # Get the min and max of the points to check

        if min_max[0] < min(self.original_data) or min_max[1] > max(
            self.original_data
        ):  # If the min or max of the points to check is outside the range of the original data
            self.x = np.linspace(  # Set the x values to the min and max of the points to check
                min(min(self.original_data), min_max[0]),
                max(max(self.original_data), min_max[1]),
                1000,
            )
            self.default_lines()  # Set the default lines
            self.widen = True  # Set widen to True
        elif self.widen:  # If widen is True
            self.x = np.linspace(
                min(self.original_data), max(self.original_data), 10000
            )  # Set the x values to the min and max of the original data
            self.default_lines()  # Set the default lines
            self.widen = False  # Set widen to False

        marks2plot = [
            self.horizontal_line,
            self.function_line,
        ]  # Set the marks to the function and the horizontal line

        # 1. The Current Points (a,b,c)
        marks2plot.append(
            draw_point(
                self.a, self.fa, "red", "a", "circle", legend=False
            )  # Draw a point
        )
        marks2plot.append(
            draw_point(
                self.b,
                self.fb,
                "Black",
                "Current Solution",
                "cross",
                legend=True,  # Draw a point
            )
        )
        marks2plot.append(
            draw_point(
                self.c, self.fc, "green", "c", "circle", legend=False
            )  # Draw a point
        )

        # 2. The next step suggestions
        marks2plot.append(  # Bisect
            draw_point(
                self.next_points_addition[0],
                0,
                "green",
                "Bisection",
                "rectangle",
            )
        )
        marks2plot.append(
            draw_point(  # Secant
                self.next_points_addition[1],
                0,
                "red",
                "Secant",
                "triangle-up",
            )
        )
        if self.next_points_addition[2] is not None:
            marks2plot.append(
                draw_point(  # IQI
                    self.next_points_addition[2],
                    0,
                    "blue",
                    "IQI",
                    "triangle-down",
                )
            )

        # FIX THE VIEW
        self.x_sc.min = min_max[
            0
        ]  # Set the x scale min to the min of the points to check
        self.x_sc.max = min_max[
            1
        ]  # Set the x scale max to the max of the points to check

        self.y_sc.min = min(
            self.f(min_max[0]), self.f(min_max[1])
        )  # Set the y scale min to the min of the function of the min and max of the points to check
        self.y_sc.max = max(
            self.f(min_max[0]), self.f(min_max[1])
        )  # Set the y scale max to the max of the function of the min and max of the points to check

        fx = list(map(self.f, self.x))  # Get the function values for the x values
        y_min_max = (min(fx), max(fx))  # Get the min and max of the function values

        # 3. The steps
        if self.secant_checkbox.value:  # If the secant checkbox is checked
            marks2plot.append(secant_draw())  # Draw the secant line

        if self.bisect_checkbox.value:  # If the bisect checkbox is checked
            marks2plot.append(  # Draw the bisect line
                bisect_draw(
                    self.next_points_addition[0], y_min_max[0], y_min_max[1]
                )  # Draw the bisect line
            )
        if (
            self.iqi_checkbox.value and self.next_points_addition[2] is not None
        ):  # If the IQI checkbox is checked and the IQI point is not None
            marks2plot.append(
                iqi_draw(self.a, self.b, self.c, y_min_max[0], y_min_max[1])
            )  # Draw the IQI line

        self.fig.marks = marks2plot  # Set the marks to the marks to plot

    def update_ouputs(self):
        """
        This method updates the outputs of the app
            > current solution
            > helper text
        The next step must be calculated before calling this method
        """
        self.current_solution_output.clear_output()  # Clear the current solution output
        self.helper_output.clear_output()  # Clear the helper output

        with self.current_solution_output:  # Print the current solution
            print(
                f"Current Solution: ({self.b:.4e}, {self.f(self.b):.4e})"
            )  # Print the current solution
            print(f"Iterations: {self.iterations}")  # Print the iterations
        with self.helper_output:  # Print the helper text
            print(  # Print the helper text
                f"Next Step suggested by Brent-Dekker: {self.hint_step if self.hint_step is not None else 'FINISHED'}"
            )

    def run(self):
        """
        This method sets up everything and runs the app

        Returns
        -------
        Widgets.GridBox
            The GridBox containing all the widgets
        """
        self.initialize_components()  # Initialize the components
        self.one_step()  # Run one step

        return self.grid

    def brent_dekker_step(self):
        """
        This method imitates the Brent-Dekker in one step, instead of giving the result point, this method returns the 3 available points for the next step as well as the next possible step

        Returns
        -------
        - next_step: str
            The next possible step
        - nextPoints_addition: list
            The 3 available quantity to add to b for the next step (midpoint, secant, IQI)
        - errs: list
            The errors of the 3 available points for the next step
        """
        self.tolerance = 2 * np.finfo(float).eps * abs(self.b) + self.t
        ## Midpoint error
        m = 0.5 * (self.c - self.b)

        if abs(m) <= self.tolerance or self.f(self.b) == 0 or self.f(self.a) == 0:
            self.hint_step = None
            self.next_points_addition = [None, None, None]
            self.errs = [None, None, None]
            self.update_ouputs()
            return

        next_step = None
        next_points_addition = [None, None, None]
        errs = [None, None, None]

        next_points_addition[0] = self.b + m  # Always exists
        errs[0] = m  # Always exists

        # See if Bisection is possible
        # CHECK abs(self.e) < self.tolerance or abs(self.fa) <= abs(self.fb) in the original code
        # Here we do nothing because it is the midpoint - only thing next_step is Bisection (which will be default if secant/iqi fails)

        s = self.fb / self.fa
        pq_pair = None
        if self.a == self.c:
            # Linear Interpolation (Secant)
            p1 = 2 * m * s
            q1 = 1 - s

        else:
            # Linear Interpolation (Secant)
            p1 = self.fb * (self.b - self.a)
            q1 = self.fb - self.fa

            # Inverse Quadratic Interpolation (IQI)
            q = self.fa / self.fc
            r = self.fb / self.fc
            p2 = s * (2 * m * q * (q - r) - (self.b - self.a) * (r - 1))
            q2 = (q - 1) * (r - 1) * (s - 1)

            # Correct signs of IQI
            if p2 > 0:
                q2 = -q2
            else:
                p2 = -p2

            pq_pair = (p2, q2)

            self.s = self.e

        if p1 > 0:
            q1 = -q1
        else:
            p1 = -p1

        # Store e-Values for next step
        errs[1] = abs(p1 / q1)  # Secant always exists
        errs[2] = (
            None if pq_pair is None else abs(pq_pair[0] / pq_pair[1])
        )  # IQI may not exist if a==c if it does, it is stored in pqPair

        # Store next points (Quantity to add) for next step
        next_points_addition[1] = self.b + p1 / q1  # Secant always exists
        next_points_addition[2] = (
            None if pq_pair is None else self.b + pq_pair[0] / pq_pair[1]
        )  # IQI may not exist if a==c if it does, it is stored in pqPair
        pq_pair = (
            (p1, q1) if pq_pair is None else pq_pair
        )  # If IQI does not exist, use Secant instead (pqPair is None if IQI does not exist), otherwise use IQI (pqPair is not None if IQI exists)

        # Choose the best interpolation
        if 2 * pq_pair[0] < 3 * m * pq_pair[1] - abs(
            self.tolerance * pq_pair[1]
        ) and pq_pair[0] < abs(0.5 * self.e * pq_pair[1]):
            next_step = "IQI" if self.a != self.c else "Secant"
        else:
            next_step = "Bisection"

        self.hint_step = next_step
        self.next_points_addition = next_points_addition
        self.errs = errs

        return next_step, next_points_addition, errs



import bqplot as bq
from ipywidgets import widgets
import numpy as np
from time import sleep
from BNumMet import Random


class RandomVisualizer:
    def __init__(self, random_generator=Random.genrand):
        """
        Initializes the RandomVisualizer class.

        Parameters
        ----------
        randGenerator : RandomGenerator
            The random generator to use. It is a function that does not take any parameters and returns a random number between 0 and 1.
        """
        self.random_generator = random_generator
        self.generated_numbers = []
        self.current_value = 0
        self.iterations = 100
        self.current_iteration = 0
        self.inside_circle = 0

    def initialize_components(self):
        """
        Initializes the components of the visualizer.
        """
        # Figure 1: Circle inside a square of side 1 centered at the origin
        # =================================================================================================
        self.figure1 = bq.Figure(title="Montecarlo's method")

        self.x_sc = bq.LinearScale()  # x scale
        self.y_sc = bq.LinearScale()  # y scale
        ax_x = bq.Axis(scale=self.x_sc, grid_lines="solid", label="X")
        ax_y = bq.Axis(
            scale=self.y_sc,
            orientation="vertical",
            grid_lines="solid",
            label="Y",
        )
        # Set up the plot figure
        self.figure1 = bq.Figure(
            marks=[],  # marks
            axes=[ax_x, ax_y],  # axes
            title="Montecarlo's Method",  # title
        )

        # Draw the square
        self.square = bq.Lines(
            x=[0, 1, 1, 0, 0],
            y=[0, 0, 1, 1, 0],
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["black"],
        )
        self.figure1.marks = [self.square]
        # Draw the full circle centered at (0.5, 0.5) with radius 0.5
        x = np.linspace(0, 1, 1000)
        y = np.sqrt(1 / 4 - (x - 0.5) ** 2)
        y = np.concatenate((y, -y))  # y
        y += 0.5
        x = np.concatenate((x, (x[::-1])))  # x

        self.circle = bq.Lines(
            x=x,
            y=y,
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["red"],
            stroke_width=5,
        )
        self.figure1.marks = [self.square, self.circle]

        # Prepare the figure for the points
        self.points = bq.Scatter(
            x=[],
            y=[],
            scales={"x": self.x_sc, "y": self.y_sc},
            colors=["blue"],
            default_size=2,
        )
        self.figure1.marks = [self.points, self.square, self.circle]

        # Figure 2: (Current value and Pi) vs. (Number of iterations)
        # =================================================================================================
        self.figure2 = bq.Figure(title="Convergence of Pi", legend_location="top-right")
        self.x_sc2 = bq.LinearScale()  # x scale
        self.y_sc2 = bq.LinearScale()  # y scale
        ax_x2 = bq.Axis(
            scale=self.x_sc2, grid_lines="solid", label="Number of iterations"
        )
        ax_y2 = bq.Axis(
            scale=self.y_sc2,
            orientation="vertical",
            grid_lines="solid",
            label="Value",
        )

        self.x_sc2.min = 0
        self.x_sc2.max = self.iterations  # Number of iterations **

        # Set up the plot figure
        self.figure2 = bq.Figure(
            marks=[],  # marks
            axes=[ax_x2, ax_y2],  # axes
            title="Convergence of Pi",  # title
        )  # this will change as the algorithm runs

        # Draw the line of Pi
        x = np.linspace(0, self.iterations, 1000)
        y = np.full(1000, np.pi)
        self.pi_line = bq.Lines(
            x=x,
            y=y,
            scales={"x": self.x_sc2, "y": self.y_sc2},
            colors=["red"],
            labels=["Pi"],
            display_legend=True,
        )
        self.figure2.marks = [self.pi_line]

        # Prepare the figure for the current value update
        self.current_value_line = bq.Lines(
            x=[],
            y=[],
            scales={"x": self.x_sc2, "y": self.y_sc2},
            colors=["blue"],
            labels=["Current Value"],
            display_legend=True,
        )
        self.figure2.marks = [self.pi_line, self.current_value_line]

        # Widget: Number of Points to generate
        # =================================================================================================
        self.iterations_widget = widgets.BoundedIntText(
            value=self.iterations,
            min=1,
            max=10000,
            description="Iterations:",
            disabled=False,
        )

        # Widget: Button for running the algorithm
        # =================================================================================================
        self.run_button = widgets.Button(
            description="Run",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Run",
            icon="check",  # (FontAwesome names without the `fa-` prefix)
        )
        # Observer: Button for running the algorithm
        self.run_button.on_click(self.play_montecarlo)

        self.pi_value = widgets.FloatText(  # Widget: Pi value
            value=0,
            description="$\pi$:",
            disabled=True,
        )
        # Widget: Grid
        # =================================================================================================
        self.grid = widgets.GridspecLayout(4, 4)
        self.grid[0:2, 0:2] = self.figure1
        self.grid[2:4, 0:4] = self.figure2
        self.grid[0:1, 2:4] = widgets.VBox([self.iterations_widget, self.run_button])

        self.grid[1:2, 2:4] = self.pi_value

    def number_of_iterations(self):
        """
        Updates the number of iterations. This is called when the user plays the animation. It configures the x axis of the second figure. and the number of iterations of the algorithm.
        """
        self.iterations = self.iterations_widget.value
        self.x_sc2.max = self.iterations
        self.pi_line.x = np.linspace(0, self.iterations, 1000)

    def play_montecarlo(self, b):
        """
        Plays the montecarlo algorithm.
        """
        self.number_of_iterations()
        self.current_iteration = 0
        self.generated_numbers = []
        self.current_value = 0
        self.inside_circle = 0
        self.points.x = []
        self.points.y = []
        self.current_value_line.x = []
        self.current_value_line.y = []
        self.total_points = 0
        self.run_button.disabled = True

        batch_coords = []
        batch_current_value = []
        batch_current_iteration = []
        batch_size = int(self.iterations / 100) * (2 if self.iterations < 2000 else 5)

        for self.current_iteration in range(self.iterations):
            new_x = self.random_generator()
            new_y = self.random_generator()
            self.total_points += 1

            batch_coords.append([new_x, new_y])
            if ((new_x - 0.5) ** 2 + (new_y - 0.5) ** 2) <= 0.25:
                self.inside_circle += 1
            batch_current_value.append(4 * self.inside_circle / self.total_points)

            batch_current_iteration.append(self.current_iteration)

            if (
                len(batch_coords) >= batch_size
                or self.current_iteration == self.iterations - 1
            ):
                aux = self.generated_numbers + batch_coords
                with self.grid.hold_sync():
                    with self.points.hold_sync():
                        self.points.x = [x[0] for x in aux]
                        self.points.y = [x[1] for x in aux]
                    with self.current_value_line.hold_sync():
                        self.current_value_line.x = (
                            list(self.current_value_line.x) + batch_current_iteration
                        )  # [x for x in self.current_value_line.x] + batch_current_iteration
                        self.current_value_line.y = (
                            list(self.current_value_line.y) + batch_current_value
                        )  # [x for x in self.current_value_line.y] + batch_current_value
                    self.pi_value.value = batch_current_value[-1]
                    self.generated_numbers += batch_coords

                batch_coords = []
                batch_current_value = []
                batch_current_iteration = []

        self.run_button.disabled = False

    def run(self):
        """
        Runs the visualizer.
        """
        self.initialize_components()
        return self.grid



import numpy as np


def lu(matrix):
    """
    LU decomposition of a square matrix A using Gaussian elimination.
    PA = LU where
        P is a permutation matrix
        L is a lower triangular matrix
        U is an upper triangular matrix

    Parameters
    ----------
    matrix : np.array
        A square matrix.

    Returns
    -------
    P : np.array
        A permutation matrix.
    L : np.array
        A lower triangular matrix.
    U : np.array
        An upper triangular matrix.

    raises
    ------
    ValueError
        If the matrix is not square.
    """

    # Check if the matrix is square
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Matrix must be square")

    n = matrix.shape[0]  # number of rows/columns
    A = matrix.copy().astype(float)  # Make a copy of A and convert to float
    P = np.eye(n, dtype=float)  # Initialize P as the identity matrix

    # Loop over the columns of A
    for col in range(n):
        # Find the index of the row with the largest pivot element
        maximum_index = int(np.argmax(np.abs(A[col:, col])) + col)

        # Skip if the pivot is zero
        if A[maximum_index, col] != 0:
            # Swap the rows of A and P for those with the largest pivot element
            A = permute(A, maximum_index, col)
            P = permute(P, maximum_index, col)

            for row in range(col + 1, n):
                A[row, col] = (
                    A[row, col] / A[col, col]
                )  # Calculate the multiplier and store in A for later use
                A[row, col + 1 :] = (
                    A[row, col + 1 :] - A[row, col] * A[col, col + 1 :]
                )  # Update the remaining elements in the row using the multiplier

    L = np.tril(A, -1) + np.eye(
        n
    )  # Extract the lower triangular matrix from A and add the identity matrix to make it a proper lower triangular matrix
    U = np.triu(A)  # Extract the upper triangular matrix from A

    return P, L, U


def interactive_lu(p, l, u, col, row, pivot_row):
    """
    Makes only one step of the LU factorization, updates P,L,U and takes the pivot row to do the Gaussian Elimination

    Parameters
    ----------
    Matrix : np.array
        A square matrix.
    P : np.array
        A permutation matrix.
    L : np.array
        A lower triangular matrix.
    U : np.array
        An upper triangular matrix.
    pivot_row : int
        The index of the pivot row.

    Returns
    -------
    P : np.array
        A permutation matrix.
    L : np.array
        A lower triangular matrix.
    U : np.array
        An upper triangular matrix.

    """

    # Check if the column index is out of bounds, if so return the original P,L,U and -1,-1 as the next column and pivot row

    # Check if the column index is out of bounds, if so return the original P,L,U and -1,-1 as the next column and pivot row
    if col >= u.shape[1] or col <= -1:
        return p, l, u, -1, row, "Finished with the LU factorization"

    # Convert the L and U matrices to float data type
    u = u.astype(float)
    l = l.astype(float)

    # Find the index of the row with the largest pivot element or use the pivot row if it is already found and the row index is greater than the pivot row
    index_maximum = (
        int(np.argmax(np.abs(u[row:, col])) + row) if row > pivot_row else pivot_row
    )
    # Skip if the pivot is zero
    if not np.isclose(u[index_maximum, col], 0):
        # Swap the rows of A and P for those with the largest pivot element
        u = permute(u, index_maximum, row)
        p = permute(p, index_maximum, row)
        l = permute(l - np.eye(l.shape[0]), index_maximum, row) + np.eye(l.shape[0])

        # Gaussian elimination using NumPy's broadcasting
        factors = u[row + 1 :, col] / u[row, col]  # Calculate the multipliers
        l[row + 1 :, col] = u[row + 1 :, col] / u[row, col]  # Update the L matrix
        for i in range(
            row + 1, u.shape[0]
        ):  # Update the remaining elements in the row using the multipliers
            u[i, col + 1 :] = u[i, col + 1 :] - factors[i - col - 1] * u[col, col + 1 :]
        u[row + 1 :, col] = 0  # Set the elements below the pivot to zero
        row += 1  # Increment the row index
    msg = ""
    while (
        row < u.shape[0] and col + 1 < u.shape[1] and np.allclose(u[row:, col + 1], 0)
    ):
        msg = "The matrix is not FULL RANK, that is, the column below the diagonal is zero therefore we move to the next column until we find a non-zero column"
        col += 1

    return (
        p,
        l,
        u,
        col + 1 if col + 1 < u.shape[1] else -1,
        row,
        msg + ("" if col + 1 < u.shape[1] else "\nWe have finished"),
    )


def permute(matrix_to_permute, i: int, j: int):
    """
    Permute the rows of a matrix.

    Parameters
    ----------
    matrix_to_permute : np.array
        A matrix.
    i : int
        The index of the first row.
    j : int
        The index of the second row.

    Returns
    -------
    matrix_to_permute : np.array
        The permuted matrix.

    """

    matrix_to_permute[[i, j], :] = matrix_to_permute[
        [j, i], :
    ]  # Swap the rows of matrix_to_permute using fancy indexing
    return matrix_to_permute


def forward_substitution(lhs, rhs):
    """
    Solves the system Ax = b using forward substitution.

    Parameters
    ----------
    lhs : np.array
        A lower triangular matrix.
    rhs : np.array
        A vector.


    Returns
    -------
    x : np.array
        A vector.

    """
    lhs = lhs.astype(float)  # Convert A to float data type
    rhs = np.copy(rhs).astype(
        float
    )  # Make a copy of b and convert it to float data type to avoid modifying the argument

    n = lhs.shape[0]  # Get the number of rows/columns in lhs
    x = np.zeros(n)  # Initialize the solution vector, x, with zeros and allocate memory

    if np.any(np.isclose(np.diag(lhs), 0, atol=1e-15)):
        # Check if the diagonal elements of A are close to zero, indicating a singular matrix
        raise ValueError("Matrix is singular")

    if np.any(np.triu(lhs, 1) != 0):
        # Check if the upper triangular part of A is not equal to zero
        raise ValueError("Matrix is not lower triangular")

    if rhs.shape[0] != n:
        # Check if the size of b is not equal to the number of rows/columns of A
        raise ValueError(
            "The size of b is not equal to the number of rows/columns of A"
        )
    if lhs.shape[0] != lhs.shape[1]:
        # Check if A is not a square matrix
        raise ValueError("A is not a square matrix")

    for row in range(n):
        # Loop over the rows starting from the first
        rhs[row] = rhs[row] - np.dot(
            lhs[row, :row], x[:row]
        )  # Subtract the dot product of the row in A and the current solution vector from the value of b
        x[row] = (
            rhs[row] / lhs[row, row]
        )  # Divide the updated b value by the diagonal element in A to get the solution for x

    return x  # Return the solution vector, x


def backward_substitution(lhs, rhs):
    """
    Solves the system Ax = b using backward substitution.

    Parameters
    ----------
    lhs : np.array
        An upper triangular matrix.
    rhs : np.array
        A vector.

    Returns
    -------
    x : np.array
        A vector.

    """
    lhs = lhs.astype(float)  # Convert A to float
    rhs = np.copy(rhs).astype(
        float
    )  # Make a copy of b (We do not want to update the argument while making our calculations) and convert to float

    n = lhs.shape[0]  # Get the number of rows/columns in the matrix
    x = np.zeros(n)  # Initialize the solution vector, x, with zeros and allocate memory

    if np.any(
        np.isclose(np.diag(lhs), 0, atol=1e-15)
    ):  # Check if the diagonal elements are close to zero (singular matrix)
        raise ValueError(
            f"Matrix is singular. The diagonal elements are {np.diag(lhs)}"
        )

    if np.any(
        np.tril(lhs, -1) != 0
    ):  # Since it is backward, we check if the lower triangular part is zero
        raise ValueError("Matrix is not upper triangular")

    if rhs.shape[0] != n:
        raise ValueError(
            "The size of b is not equal to the number of rows/columns of A"
        )
    if lhs.shape[0] != lhs.shape[1]:
        raise ValueError("A is not a square matrix")

    for row in range(n - 1, -1, -1):  # Loop over the rows starting from the last
        rhs[row] = rhs[row] - np.dot(
            lhs[row, row + 1 :], x[row + 1 :]
        )  # Calculate the value of b with <A[row, row+1:], x[row+1:]> subtracted (b-<A[row, row+1:], x[row+1:]>)
        x[row] = rhs[row] / lhs[row, row]  # x = b'/A[row, row]

    return x


def lu_solve(A, b):
    """
    Solves the system Ax = b using LU factorization.

    Parameters
    ----------
    A : np.array
        A square matrix.
    b : np.array
        A vector.

    Returns
    -------
    x : np.array
        A vector.

    """
    P, L, U = lu(A)  # LU factorization
    y = forward_substitution(L, P @ b)  # Solve Ly = Pb
    x = backward_substitution(U, y)  # Solve Ux = y

    return x


def qr_factorization(A):
    """
    QR using Householder reflections.

    Parameters
    ----------
    A : np.array
        A matrix.

    Returns
    -------
    Q : np.array
        An orthogonal matrix.
    R : np.array
        An upper triangular matrix.

    """
    m, n = A.shape  # Get the shape of the input matrix A
    R = A.copy().astype(float)  # Make a copy of A and convert it to float
    qs = np.zeros((m, n))  # Create an array of zeros with shape (m, n)

    for k in range(n):  # Loop through the columns of A
        x = R[k:, k]  # Get the k-th column of R, starting from the k-th row

        # Calculate the Householder vector:
        # 1. Take the sign of the first element of x
        # 2. Compute the norm of x
        # 3. Create an identity matrix with the same number of rows as x
        # 4. Add the result of step 1 to the identity matrix, scaled by the result of step 2
        qk = np.sign(x[0]) * np.linalg.norm(x) * np.eye(x.shape[0], 1) + x.reshape(
            -1, 1
        )

        # Normalize the Householder vector
        qk = qk / np.linalg.norm(qk)

        # Update the k-th column and the columns below it of R
        R[k:, k:] = R[k:, k:] - 2 * np.dot(qk, np.dot(qk.T, R[k:, k:]))

        # Store the Householder vector in the "qs" matrix
        qs[k:, k] = qk.reshape(-1)

    Q = np.eye(m)  # Create an identity matrix with shape (m, m)
    for k in range(n - 1, -1, -1):  # Loop through the columns of A in reverse order
        qk = qs[k:, k].reshape(-1, 1)  # Get the k-th column of "qs"

        # Update the k-th column and the columns below it of Q
        Q[k:, k:] = Q[k:, k:] - 2 * np.dot(qk, np.dot(qk.T, Q[k:, k:]))
    R = np.triu(R)  # Make R upper triangular
    return Q, R  # Return the matrices Q and R


def qr_solve(A, b):
    """
    Solves the system Ax = b using QR factorization without calculating Q, only R. This is faster than the QR factorization with Q.

    Parameters
    ----------
    A : np.array
        A square matrix.
    b : np.array
        A vector.

    Returns
    -------
    x : np.array
        A vector.

    """
    n = A.shape[1]  # Get the number of rows in A

    R = A.copy().astype(float)  # Create a copy of A and cast it as a float
    b = b.copy().astype(float)  # Create a copy of b and cast it as a float

    for k in range(n):
        x = R[k:, k]  # Get the kth column of R, starting from the kth row
        qk = np.sign(x[0]) * np.linalg.norm(x) * np.eye(x.shape[0], 1) + x.reshape(
            -1, 1
        )  # Householder vector: calculate the Householder reflection vector qk
        qk = qk / np.linalg.norm(qk)  # Normalize qk so it has a length of 1

        R[k:, k:] = R[k:, k:] - 2 * np.dot(
            qk, np.dot(qk.T, R[k:, k:])
        )  # Update the R matrix using qk
        b[k:] = b[k:] - 2 * np.dot(qk, np.dot(qk.T, b[k:]))  # Update b using qk

    R = np.triu(
        R
    )  # Make the lower triangular part of R zero, since it may contain small values due to numerical errors
    x = backward_substitution(
        R[:n, :n], b[:n]
    )  # Solve the system R*x = b using backward substitution

    return x  # Return the solution x



from BNumMet import Random
from unittest import TestCase
import numpy as np
from BNumMet.Visualizers.RandomVisualizer import RandomVisualizer


class test_Random(TestCase):
    test_label = "Random Number Generator"

    def test_lehmers_start(self):
        Random.clear_lehmers_vars()
        # Tests that Lehmers Variables are initialized to None
        self.assertEqual(Random.lehmers_vars["a"], None)
        self.assertEqual(Random.lehmers_vars["c"], None)
        self.assertEqual(Random.lehmers_vars["m"], None)
        self.assertEqual(Random.lehmers_vars["x"], None)

    def test_lehmers_init(self):
        # Tests that Lehmers Variables are initialized correctly
        Random.clear_lehmers_vars()
        Random.lehmers_init(2, 3, 5, 1)
        self.assertEqual(Random.lehmers_vars["a"], 2)
        self.assertEqual(Random.lehmers_vars["c"], 3)
        self.assertEqual(Random.lehmers_vars["m"], 5)
        self.assertEqual(Random.lehmers_vars["x"], 1)
        Random.clear_lehmers_vars()

    def test_lehmers_default(self):
        # Tests that Lehmers Variables are initialized to default values
        Random.clear_lehmers_vars()
        Random.lehmers_rand()
        self.assertEqual(Random.lehmers_vars["a"], 7**5)
        self.assertEqual(Random.lehmers_vars["c"], 0)
        self.assertEqual(Random.lehmers_vars["m"], 2**31 - 1)
        Random.clear_lehmers_vars()

    def test_lehmers_initFixed(self):
        # Tests that Lehmers Variables are initialized correctly
        Random.clear_lehmers_vars()
        Random.lehmers_rand(2, 3, 5, 1)
        self.assertEqual(Random.lehmers_vars["a"], 2)
        self.assertEqual(Random.lehmers_vars["c"], 3)
        self.assertEqual(Random.lehmers_vars["m"], 5)
        Random.clear_lehmers_vars()

    def test_lehmers_rand_formula(self):
        arr = [1]
        maxIter = 100
        for i in range(maxIter):
            aux = Random.lehmers_rand(a=2**16 + 3, m=2**31, c=0, x=arr[-1])
            if len(arr) >= 3:
                lehmerFormula = (
                    6 * arr[-1] - 9 * arr[-2]
                ) % 1  # Test Xn = (6Xn-1 - 9Xn-2)
                self.assertEqual(lehmerFormula, aux)
            arr.append(aux)

    def test_marsaglia_start(self):
        Random.clear_marsaglia_vars()
        # Tests that Marsaglia Variables are initialized to None
        self.assertEqual(Random.marsaglia_vars["base"], None)
        self.assertEqual(Random.marsaglia_vars["lag_r"], None)
        self.assertEqual(Random.marsaglia_vars["lag_s"], None)
        self.assertEqual(Random.marsaglia_vars["carry"], None)
        self.assertEqual(Random.marsaglia_vars["args"], None)

    def test_marsaglia_init(self):
        # Tests that Marsaglia Variables are initialized correctly
        Random.clear_marsaglia_vars()
        Random.marsaglia_init(2, 2, 1, 1, seed_tuple=(1, 2))
        self.assertEqual(Random.marsaglia_vars["base"], 2)
        self.assertEqual(Random.marsaglia_vars["lag_r"], 2)
        self.assertEqual(Random.marsaglia_vars["lag_s"], 1)
        self.assertEqual(Random.marsaglia_vars["carry"], 1)
        self.assertEqual(Random.marsaglia_vars["args"], [1, 2])

        Random.clear_marsaglia_vars()
        # ValueError: seedTuple must be a tuple of length 2
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 3, 5, 1, seed_tuple=1)
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 3, 5, 1, seed_tuple=(1, 2, 3))
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 3, 5, 1, seed_tuple=(1,))

        # ValueError: lag_r and lag_s must be greater than 0
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 0, 5, 1, seed_tuple=(1, 2))
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 3, 0, 1, seed_tuple=(1, 2))

        # ValueError: lag_r must be greater than or equal to lag_s
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 2, 3, 1, seed_tuple=(1, 2))

        # ValueError: carry must be 0 or 1
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 5, 3, 24, seed_tuple=(1, 2))

        # ValueError: base must be greater than 0
        with self.assertRaises(ValueError):
            Random.marsaglia_init(0, 5, 3, 1, seed_tuple=(1, 2))

        # Check if lag_r and lag_s are greater than 0
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, -1, 5, 1, seed_tuple=(1, 2))
        with self.assertRaises(ValueError):
            Random.marsaglia_init(2, 3, 0, 1, seed_tuple=(1, 2))

    def test_marsaglia_default(self):
        # Tests that Marsaglia Variables are initialized to default values
        Random.clear_marsaglia_vars()
        Random.marsaglia_rand()
        self.assertEqual(Random.marsaglia_vars["base"], 2**31 - 1)
        self.assertEqual(Random.marsaglia_vars["lag_r"], 19)
        self.assertEqual(Random.marsaglia_vars["lag_s"], 7)
        self.assertEqual(Random.marsaglia_vars["carry"], 1)
        Random.clear_marsaglia_vars()

    def test_marsaglia_initFixed(self):
        # Tests that Marsaglia Variables are initialized correctly
        Random.clear_marsaglia_vars()
        Random.marsaglia_rand(2, 2, 1, 1, seed_tuple=(1, 2))
        self.assertEqual(Random.marsaglia_vars["base"], 2)
        self.assertEqual(Random.marsaglia_vars["lag_r"], 2)
        self.assertEqual(Random.marsaglia_vars["lag_s"], 1)
        self.assertEqual(Random.marsaglia_vars["carry"], 1)
        self.assertEqual(Random.marsaglia_vars["args"][0], 2)
        Random.clear_marsaglia_vars()

    def test_marsaglia_rand_formula(self):
        testArr = [
            0.0,
            1,
            0.9,
            0.1,
            0.7,
            0.4,
            0.2,
            0.2,
            0.0,
            0.2,
            0.8,
            0.3,
            0.4,
            0.9,
        ]  # According to Bibliography
        resArr = [0, 1]
        Random.clear_marsaglia_vars()
        for _ in range(len(testArr) - 2):
            resArr.append(
                Random.marsaglia_rand(
                    base=10, lag_r=2, lag_s=1, carry=0, seed_tuple=(0, 1)
                )
            )
            self.assertEqual(testArr[: len(resArr)], resArr)

    def test_mt_start(self):
        Random.clear_mt_vars()
        # Tests that Mersenne Variables are initialized to None
        mt_vars = {
            "N": None,  # N
            "M": None,  # M
            "MATRIX_A": None,  # MATRIX_A
            "UPPER_MASK": None,  # UPPER_MASK
            "LOWER_MASK": None,  # LOWER_MASK
            "TEMPERING_MASK_B": None,  # TEMPERING_MASK_B
            "TEMPERING_MASK_C": None,  # TEMPERING_MASK_C
            "mt": None,  # mt
            "mti": None,  # mti
        }
        self.assertEqual(Random.mt_vars, mt_vars)

        Random.clear_mt_vars()
        # Tests that Mersenne Variables are initialized correctly
        Random.sgenrand(1)
        self.assertEqual(Random.mt_vars["N"], 624)
        self.assertEqual(Random.mt_vars["M"], 397)
        self.assertEqual(Random.mt_vars["MATRIX_A"], 0x9908B0DF)
        self.assertEqual(Random.mt_vars["UPPER_MASK"], 0x80000000)
        self.assertEqual(Random.mt_vars["LOWER_MASK"], 0x7FFFFFFF)
        self.assertEqual(Random.mt_vars["TEMPERING_MASK_B"], 0x9D2C5680)
        self.assertEqual(Random.mt_vars["TEMPERING_MASK_C"], 0xEFC60000)
        self.assertEqual(Random.mt_vars["mt"][0], 1 & 0xFFFFFFFF)
        self.assertEqual(Random.mt_vars["mti"], 624)  # mti

    def test_mt_init(self):
        Random.clear_mt_vars()
        # Tests that Mersenne Variables are initialized correctly
        Random.genrand(1)
        self.assertEqual(Random.mt_vars["N"], 624)
        self.assertEqual(Random.mt_vars["M"], 397)
        self.assertEqual(Random.mt_vars["MATRIX_A"], 0x9908B0DF)
        self.assertEqual(Random.mt_vars["UPPER_MASK"], 0x80000000)
        self.assertEqual(Random.mt_vars["LOWER_MASK"], 0x7FFFFFFF)
        self.assertEqual(Random.mt_vars["TEMPERING_MASK_B"], 0x9D2C5680)
        self.assertEqual(Random.mt_vars["TEMPERING_MASK_C"], 0xEFC60000)
        self.assertNotEqual(
            Random.mt_vars["mt"][0], 1 & 0xFFFFFFFF
        )  # Not equal because it has executed the generator
        self.assertEqual(
            Random.mt_vars["mti"], 1
        )  # Must be 1 because it has executed the generator

        Random.clear_mt_vars()
        # Initialize not with Int
        with self.assertRaises(ValueError):
            Random.genrand("1")

    def test_mt_randomness(self):
        from nistrng import (
            pack_sequence,
            check_eligibility_all_battery,
            run_all_battery,
            SP800_22R1A_BATTERY,
        )

        Random.clear_mt_vars()

        sequence = np.array(
            [Random.genrand() * 0xFFFFFFFF for i in range(100)], dtype=np.uint64
        )

        binary_sequence: np.ndarray = pack_sequence(sequence)

        # Check the eligibility of the test and generate an eligible battery from the default NIST-sp800-22r1a battery
        eligible_battery: dict = check_eligibility_all_battery(
            binary_sequence, SP800_22R1A_BATTERY
        )

        # Test the sequence on the eligible tests
        results = run_all_battery(binary_sequence, eligible_battery, False)
        # Print results one by one
        results = [result.passed for result, _ in results if result.passed]

        self.assertGreaterEqual(len(results), 9)


class test_RandomVisualizer(TestCase):
    def test_initiate(self):
        # Tests that the RandomVisualizer is initiated correctly without any arguments
        # No Exceptions should be raised
        rv = RandomVisualizer()

        self.assertTrue(len(rv.generated_numbers) == 0)
        self.assertEqual(rv.iterations, 100)

        random_float = rv.random_generator()
        self.assertGreaterEqual(random_float, 0)
        self.assertLessEqual(random_float, 1)

    def test_initiate_with_args(self):
        # Tests that the RandomVisualizer is initiated correctly with arguments
        # No Exceptions should be raised
        rv = RandomVisualizer(random_generator=Random.marsaglia_rand)

        self.assertTrue(len(rv.generated_numbers) == 0)
        self.assertEqual(rv.iterations, 100)

        random_float = rv.random_generator()
        self.assertGreaterEqual(random_float, 0)
        self.assertLessEqual(random_float, 1)

    def test_run_no_exception(self):
        # Tests that the RandomVisualizer is initiated correctly without any arguments
        # No Exceptions should be raised
        rv = RandomVisualizer()
        rv.run()

    def test_play(self):
        # Tests that the RandomVisualizer is initiated correctly without any arguments
        # No Exceptions should be raised
        rv = RandomVisualizer()
        rv.run()
        rv.run_button.click()

        self.assertTrue(rv.current_value - np.pi < 0.7)

    def test_int_text(self):
        # Tests that the RandomVisualizer is initiated correctly without any arguments
        # No Exceptions should be raised
        rv = RandomVisualizer()
        rv.run()

        rv.iterations_widget.value = 10000
        self.assertEqual(
            rv.iterations_widget.value, 10000
        )  # Should be 1000 because of the max value

        rv.iterations_widget.value = 0
        self.assertEqual(
            rv.iterations_widget.value, 1
        )  # Should be 1 because of the min value
